/*
 * File: htext_common.c
 * Common routines for all handlers
 */
#define _GNU_SOURCE
#include <ctype.h>
#include <curl/curl.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "htext.h"
#include "htext_private.h"

/**
 * Initializes a previously allocated partial
 */
void htext_partial_init(htext_partial *p)
{
  memset(p, 0, sizeof(htext_partial));
  sem_init(&(p->final), 0, 0);
}

/**
 * Cleans a partial, but does NOT free it
 */
void htext_partial_clean(htext_partial *p)
{
  if (p->curl) curl_easy_cleanup(p->curl);
  if (p->fd)   fclose(p->fd);
  free(p->location);
  if (p->headers) curl_slist_free_all(p->headers);
  sem_destroy(&(p->final));
  free(p->delegation_service);
}

/**
 * Removes all whitespaces at the beginning and end
 * @param The string to trim
 * @return str
 */
char *trim(char *str)
{
  int i, j;
  /* Remove at the beginning */
  for (i = 0; str[i] != '\0' && isspace(str[i]); ++i);
  for (j = i, i = 0; str[j] != '\0'; ++i, ++j)
    str[i] = str[j];
  /* Remove at the end */
  for (--j; j >= 0 && isspace(str[j]); --j)
    str[j] = '\0';
  return str;
}

/**
 * Puts in the buffer the host extracted from url
 * @param url    The url to parse
 * @param buffer Where to put the host
 * @return buffer
 */
char *get_host(const char *url, char *buffer)
{
  char *start, *end;
  int   i;

  start = index(url, ':') + 3; /* Jump over ://     */
  end   = index(start, '/');   /* If NULL, it is ok */

  for (i = 0; start[i] != '\0' && start + i != end; ++i)
    buffer[i] = start[i];
  buffer[i] = '\0';
  return buffer;
}

/**
 * Sends a log entry to the callback
 * @param handle The handle
 * @param fmt    A format string
 */
int htext_log(htext_handle *handle, const char *fmt, ...)
{
  char               buffer[512];
  int                n;
  va_list            args;
  htext_log_callback f;

  va_start(args, fmt);
  n = vsnprintf(buffer, sizeof(buffer), fmt, args);
  va_end(args);

  f = (htext_log_callback)GETPTR(handle, HTEXTOP_LOGCALLBACK);
  if (f)
    f(handle, HTEXT_LOG, buffer, strlen(buffer), GETPTR(handle,HTEXTOP_LOGCALLBACK_DATA));

  return n;
}


/**
 * Sets the error string
 * @param handle The handle
 * @param fmt    A format string
 */
int htext_error(htext_handle *handle, const char *fmt, ...)
{
  char               buffer[512];
  int                n;
  va_list            args;

  if (fmt != NULL) {
    va_start(args, fmt);
    n = vsnprintf(buffer, sizeof(buffer), fmt, args);
    va_end(args);

    handle->error_string = strdup(buffer);
  }
  handle->status = HTEXTS_FAILED;

  return n;
}

/**
 * Called by Curl every time a line of the header needs to be processed
 * @param buffer   Where the data is
 * @param size     The size of the element
 * @param nmemb    The number of elements
 * @param st       A pointer to a htext_partial structure
 * @return         The read size
 */
size_t htext_header_callback(void *buffer, size_t size, size_t nmemb, void *st)
{
  htext_partial *partial = (htext_partial*)st;
  size_t         bsize   = size * nmemb;
  char          *line    = (char*)buffer;
  char          *p;

  /* Do nothing if partial not set */
  if (!partial)
    return bsize;

  /* HTTP status code */
  if (strncasecmp("HTTP/", line, 5) == 0) {
    p = index(line, ' ') + 1;
    partial->http_status = atoi(p);
    /* Check it is no redirect */
    if ((partial->http_status < 300 || partial->http_status >= 400)) {
      sem_post(&(partial->final));
    }
    /* It is, so reset some fields */
    else {
      free(partial->delegation_service);
      switch (partial->http_status) {
        case 300: case 301:
          partial->redirect_is_cacheable = 1;
          break;
        default:
          partial->redirect_is_cacheable = 0;
      }
    }
  }
  /* Actual location of the file */
  else if (strncasecmp("Location:", line, 9) == 0 || strncasecmp("Content-Location:", line, 17) == 0) {
    p = index(line, ':') + 1;
    partial->location = trim(strdup(p));
  }
  /* File size */
  else if (strncasecmp("Content-Length:", line, 15) == 0) {
    p = index(line, ':') + 1;
    *(partial->partial_total) = atol(p);
  }
  /* Chunked transfer */
  else if (strncasecmp("Transfer-Encoding: chunked", line, 27) == 0) {
    *(partial->partial_total) = 0; /* Who knows? */
  }
  /* Delegation service */
  else if (strncasecmp("Proxy-Delegation-Service:", line, 25) == 0) {
    p = index(line, ':') + 1;
    partial->delegation_service = trim(strdup(p));
  }
  /* Caching */
  else if (strncasecmp("Cache-Control:", line, 14) == 0) {
    p = index(line, ':') + 1;
    p = trim(p);
    if (strcasecmp("no-cache", p) == 0)
      partial->redirect_is_cacheable = 0;
  }
  else if (strncasecmp("Expires:", line, 8) == 0) {
    struct tm st_time;
    p = index(line, ':') + 1;
    p = trim(p);
    /* Scan the specified date */
    if (strptime(p, "%a, %d %b %Y %H:%M:%S GMT", &st_time) == NULL)
      partial->redirect_is_cacheable = 0;
    else if(difftime(mktime(&st_time), time(NULL)) <= 0)
      partial->redirect_is_cacheable = 0;
    else
      partial->redirect_is_cacheable = 1;
  }

  return bsize;
}

/**
 * Called to keep track of the progress
 * @param pp       A pointer to a htext_partial structure
 * @param dltotal  To be downloaded
 * @param dlnow    Already downloaded
 * @param ultotal  To be uploaded
 * @param ulnow    Already uploaded
 * @return         0 on success
 */
int htext_progress_callback(void *pp, double dltotal, double dlnow,
                                      double ultotal, double ulnow)
{
  htext_partial *partial = (htext_partial*)pp;
  if (partial) {
    if (partial->handle->http_method == HTTP_PUT) {
      *(partial->partial_total) = ultotal;
      *(partial->partial_done)  = ulnow;
    }
    else {
      *(partial->partial_total) = dltotal;
      *(partial->partial_done)  = dlnow;
    }
  }
  return 0;
}

/**
 * Method called by Curl to write received data from the server
 * @param buffer The retrieved data
 * @param size   The size of an element
 * @param nmemb  The number of elements
 * @param pp     A pointer to a htext_partial structure
 * @return       The number of written bytes
 */
size_t htext_write_callback(char *buffer, size_t size, size_t nmemb, void *pp)
{
  htext_partial *partial = (htext_partial*)pp;
  size_t         bsize   = size * nmemb;

  if (GETINT(partial->handle, HTEXTOP_VERBOSITY) > 0 && bsize) {
    htext_log_callback f = (htext_log_callback)GETPTR(partial->handle, HTEXTOP_LOGCALLBACK);
    f(partial->handle, HTEXT_BODY, buffer, bsize, GETPTR(partial->handle, HTEXTOP_LOGCALLBACK_DATA));
  }

  return bsize;
}

/**
 * Method called by Curl to write received data from the server, dummy implementation
 * @param buffer The retrieved data
 * @param size   The size of an element
 * @param nmemb  The number of elements
 * @param pp     A pointer to a htext_partial structure
 * @return       The number of written bytes
 */
size_t htext_dummy_write_callback(char *buffer, size_t size, size_t nmemb, void *pp)
{
  /* We want to ignore this output */
  return size * nmemb;
}

/**
 * Method called by Curl on debug output
 * @param curl The CURL handle calling
 * @param type The type of data passed
 * @param data The debug information
 * @param size The size of the buffer passed in data
 * @param pp   A pointer to a htext_partial structure
 * @return     0 on success
 */
int htext_debug_callback(CURL *curl, curl_infotype type, char *data, size_t size, void *pp)
{
  htext_partial *partial = (htext_partial*)pp;

  if (partial && GETINT(partial->handle, HTEXTOP_VERBOSITY) > 1) {
    htext_log_callback  f = (htext_log_callback)GETPTR(partial->handle, HTEXTOP_LOGCALLBACK);

    switch (type) {
      case CURLINFO_HEADER_IN:
        f(partial->handle, HTEXT_HEADER_IN, data, size, GETPTR(partial->handle, HTEXTOP_LOGCALLBACK_DATA));
        break;
      case CURLINFO_HEADER_OUT:
        f(partial->handle, HTEXT_HEADER_OUT, data, size, GETPTR(partial->handle, HTEXTOP_LOGCALLBACK_DATA));
        break;
      default:
        break;
    }
  }
  
  return 0;
}
