#include "mod_dav_dpm.h"
#include <apr_strings.h>
#include <apr_time.h>
#include <http_log.h>
#include <time.h>

/* Internal structure for supported properties */
struct prop_handler {
  const char *ns;
  const char *name;
  char *(*handler)(dav_db *db);
};

/* Internal structure for property DB handling */
struct dav_db {
  /* Context */
  apr_pool_t         *pool;
  const dav_resource *resource;
  const request_rec  *request;
  /* Read-only? */
  int ro;
  /* Used to iterate by dav_dpm_propdb_*name */
  struct prop_handler *iter;
};

/**
 * Open the properties DB
 * @param p The pool used to allocate memory
 * @param resource The resource being processed
 * @param ro       Read-only mode (1)
 * @param pdb      Here we put the created dav_db
 * @return         NULL on success
 */
static dav_error *dav_dpm_propdb_open(apr_pool_t *p, const dav_resource *resource,
                                      int ro, dav_db **pdb)
{
  dav_db *db;

  /* Dead properties not supported yet (for read only, we pretend, so mod_dav works) */
  if (!ro)
    return dav_dpm_new_error(p, ENOSYS, 0, "Write mode for properties not supported");

  /* Initialize */
  db = apr_pcalloc(p, sizeof(dav_db));
  db->pool     = p;
  db->resource = resource;
  db->request  = resource->info->request;
  db->ro       = ro;

  db->iter     = NULL;

  *pdb = db;

  return NULL;
}

/**
 * Closes the property handler
 * @param db The DB to close
 */
static void dav_dpm_propdb_close(dav_db *db)
{
  /* Nothing to do here */
}

/**
 * Define namespaces the values may need
 * @param db
 * @param xi
 * @return
 */
static dav_error *dav_dpm_propdb_define_namespaces(dav_db *db,
                                                   dav_xmlns_info *xi)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_define_namespaces not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_propdb_define_namespaces");
}

/**
 * Put the value for the specified property into phdr
 * @param db    The database being examinated
 * @param name  The property name
 * @param xi    Namespace handling
 * @param phdr  Where to attach the value
 * @param found Must be set to 1 if the attribute is found
 * @return      NULL on success
 */
static dav_error *dav_dpm_propdb_output_value(dav_db *db, const dav_prop_name *name,
                                              dav_xmlns_info *xi,
                                              apr_text_header *phdr, int *found)
{
  /* Ignore */
  return NULL;
}

/**
 * Build a mapping from "global" namespaces into provider-local namespace identifiers.
 * This is done before calling store hook
 * @param db
 * @param namespaces
 * @param mapping
 * @return
 */
static dav_error *dav_dpm_propdb_map_namespaces(dav_db *db,
                                                const apr_array_header_t *namespaces,
                                                dav_namespace_map **mapping)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_map_namespaces not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_propdb_map_namespaces");
}

/**
 * Store a property value
 * @param db
 * @param name
 * @param elem
 * @param mapping
 * @return
 */
static dav_error *dav_dpm_propdb_store(dav_db *db, const dav_prop_name *name,
                                       const apr_xml_elem *elem,
                                       dav_namespace_map *mapping)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_store not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_propdb_store");
}

/**
 * Remove a property
 * @param db
 * @param name
 * @return
 */
static dav_error *dav_dpm_propdb_remove(dav_db *db, const dav_prop_name *name)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_remove not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_propdb_remove");
}

/**
 * Check if a given property is already defined
 * @param db
 * @param name
 * @return
 */
static int dav_dpm_propdb_exists(dav_db *db, const dav_prop_name *name)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_exists not implemented (%s:%s)", name->ns, name->name);
  return 0;
}

/**
 * Return the first supported property
 * @param db    The active DB instance
 * @param pname Where to put the property
 * @return      NULL on success
 */
static dav_error *dav_dpm_propdb_first_name(dav_db *db, dav_prop_name *pname)
{
  /* Not supported. mod_dav is picky, though, and we will have trouble if this
   * returns a "not-implemented" error */
  pname->ns = pname->name = NULL;
  return NULL;
}

/**
 * Iterates to the next supported property
 * @param db    The active DB instance
 * @param pname Where to put the property
 * @return      NULL on success
 */
static dav_error *dav_dpm_propdb_next_name(dav_db *db, dav_prop_name *pname)
{
  /* Not supported. mod_dav is picky, though, and we will have trouble if this
   * returns a "not-implemented" error */
  pname->ns = pname->name = NULL;
  return NULL;
}

/**
 * Get rollback context
 * @param db
 * @param name
 * @param prollback
 * @return
 */
static dav_error *dav_dpm_propdb_get_rollback(dav_db *db, const dav_prop_name *name,
                                              dav_deadprop_rollback **prollback)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_get_rollback not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_propdb_get_rollback");
}

/**
 * Apply rollback
 * @param db
 * @param rollback
 * @return
 */
static dav_error *dav_dpm_propdb_apply_rollback(dav_db *db,
                                                dav_deadprop_rollback *rollback)
{
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                db->request,
                "dav_dpm_propdb_apply_rollback not implemented");
  return dav_dpm_new_error(db->pool, ENOSYS, 0, "dav_dpm_apply_rollback");
}

/* Properties hooks */
const dav_hooks_db dav_hooks_db_dbm = {
    dav_dpm_propdb_open,
    dav_dpm_propdb_close,
    dav_dpm_propdb_define_namespaces,
    dav_dpm_propdb_output_value,
    dav_dpm_propdb_map_namespaces,
    dav_dpm_propdb_store,
    dav_dpm_propdb_remove,
    dav_dpm_propdb_exists,
    dav_dpm_propdb_first_name,
    dav_dpm_propdb_next_name,
    dav_dpm_propdb_get_rollback,
    dav_dpm_propdb_apply_rollback,
    NULL  /* Context */
};
