#define _GNU_SOURCE
#include "mod_dav_dpm.h"
#include <apr_strings.h>
#include <ctype.h>
#include <Cns_api.h>
#include <http_log.h>
#include <inttypes.h>
#include <serrno.h>
#include <sys/stat.h>

#include "replicas.h"

/* Not defined in mod_dav.h */
enum {
  /* Non-standard */
  DAV_PROPID_executable,
  DAV_PROPID_executable_apache,
  DAV_PROPID_iscollection,
  /* LCGDM specific */
  DAV_PROPID_guid,
  DAV_PROPID_mode,
  DAV_PROPID_sumtype,
  DAV_PROPID_sumvalue,
  DAV_PROPID_replicas,
  DAV_PROPID_fileid,
  DAV_PROPID_fileclass,
  DAV_PROPID_status,
  DAV_PROPID_lastaccess,
  /* Defined by http://webdav.org/specs/rfc3744.html (ACL)*/
  DAV_PROPID_owner,
  DAV_PROPID_group,
};

/*
 * Namespace definitions
 * So far, only default
 */
static const char * const dav_dpm_namespace_uris[] =
{
  "DAV:",
  "http://apache.org/dav/props/",
  "LCGDM:",
  NULL
};
enum {
  DAV_URI_DAV,
  DAV_URI_APACHE,
  DAV_URI_LCGDM
};

/* Used for the property context */
typedef union {
  char    flag;
  uid_t   uid;
  gid_t   gid;
  mode_t  mode;
  char   *string;
  time_t  tim;
  struct dav_ns_replica_array replica_array;
} dav_dpm_liveprop_context;

/* Properties definition */
static const dav_liveprop_spec dav_dpm_props[] =
{
    /* Standard DAV properties */
    {
      DAV_URI_DAV,
      "creationdate",
      DAV_PROPID_creationdate,
      0
    },
    {
      DAV_URI_DAV,
      "getlastmodified",
      DAV_PROPID_getlastmodified,
      1
    },
    {
      DAV_URI_LCGDM,
      "lastaccessed",
      DAV_PROPID_lastaccess,
      1
    },
    {
      DAV_URI_DAV,
      "getetag",
      DAV_PROPID_getetag,
      0
    },
    {
      DAV_URI_DAV,
      "getcontentlength",
      DAV_PROPID_getcontentlength,
      0
    },
    {
      DAV_URI_DAV,
      "displayname",
      DAV_PROPID_displayname,
      0
    },
    {
      DAV_URI_DAV,
      "getcontenttype",
      DAV_PROPID_getcontenttype,
      0
    },
    /* Standard from DeltaV */
    {
      DAV_URI_DAV,
      "comment",
      DAV_PROPID_comment,
      1
    },
    /* Non standard */
    {
      DAV_URI_DAV,
      "executable",
      DAV_PROPID_executable,
      1
    },
    {
      DAV_URI_APACHE,
      "executable",
      DAV_PROPID_executable_apache,
      1
    },
    {
      DAV_URI_DAV,
      "iscollection",
      DAV_PROPID_iscollection,
      0
    },
    /* LCGDM specific */
    {
      DAV_URI_LCGDM,
      "guid",
      DAV_PROPID_guid,
      0
    },
    {
      DAV_URI_LCGDM,
      "mode",
      DAV_PROPID_mode,
      1
    },
    {
      DAV_URI_LCGDM,
      "sumtype",
      DAV_PROPID_sumtype,
      0
    },
    {
      DAV_URI_LCGDM,
      "sumvalue",
      DAV_PROPID_sumvalue,
      0
    },
    {
      DAV_URI_LCGDM,
      "replicas",
      DAV_PROPID_replicas,
      1
    },
    {
      DAV_URI_LCGDM,
      "fileid",
      DAV_PROPID_fileid,
      0
    },
    {
      DAV_URI_LCGDM,
      "fileclass",
      DAV_PROPID_fileclass,
      0
    },
    {
      DAV_URI_LCGDM,
      "status",
      DAV_PROPID_status,
      0
    },
    /* ACL RFC  */
    {
      DAV_URI_DAV,
      "owner",
      DAV_PROPID_owner,
      1
    },
    {
      DAV_URI_DAV,
      "group",
      DAV_PROPID_group,
      1
    },
    /* END */
    {0}
};

static char has_group(const dav_dpm_user_info *user, gid_t gid)
{
  int i;
  for (i = 0; i < user->nfqans; ++i) {
    if (user->gids[i] == gid)
      return 1;
  }
  return 0;
}

/**
 * Insert a property into the output (retrieve property)
 * @param resource The resource being queried
 * @param propid   The property ID
 * @param what     What to get
 * @param phdr     Where to put
 * @return         Operation status
 */
static dav_prop_insert dav_dpm_insert_prop(const dav_resource *resource,
                                           int propid, dav_prop_insert what,
                                           apr_text_header *phdr)
{
  const char              *p, *p2;
  char                    *s;
  char                     buffer[CA_MAXSFNLEN + 1]; /* CA_MAXSFNLEN > CA_MAXCOMMENTLEN */
  char                    *value = buffer;
  int                      global_ns, nreplies, i;
  const dav_liveprop_spec *prop_info;
  struct Cns_filereplica  *replicas;

  /* No existence, no properties */
  if (!resource->exists)
    return DAV_PROP_INSERT_NOTDEF;

  /* Gimme! */
  switch(propid) {

    /* Creation date, RFC3339 expected */
    case DAV_PROPID_creationdate:
      dav_dpm_format_datetime(buffer, sizeof(buffer), resource->info->stat.mtime, DAV_DPM_RFC3339);
      break;

    /* Last modification date, RFC1123 expected */
    case DAV_PROPID_getlastmodified:
      dav_dpm_format_datetime(buffer, sizeof(buffer), resource->info->stat.mtime, DAV_DPM_RFC2068);
      break;

    /* Last access time */
    case DAV_PROPID_lastaccess:
      dav_dpm_format_datetime(buffer, sizeof(buffer),
                              resource->info->stat.atime, DAV_DPM_RFC2068);
      break;

    /* Size of the file. Number of entries for directories */
    case DAV_PROPID_getcontentlength:
      if (resource->collection)
        snprintf(buffer, sizeof(buffer), "%u", resource->info->stat.nlink);
      else
        snprintf(buffer, sizeof(buffer), "%llu", resource->info->stat.filesize);
      break;

    /* 1 if the resource is a collection. Used by Windows Explorer */
    case DAV_PROPID_iscollection:
      value[0] = resource->collection ? '1':'0';
      value[1] = '\0';
      break;

    /* Return 'T' if the file is executable */
    case DAV_PROPID_executable: case DAV_PROPID_executable_apache:
      if (resource->collection)
        return DAV_PROP_INSERT_NOTDEF;
      else
        value[0] = (resource->info->stat.filemode & S_IXUSR) ? 'T':'F';
        value[1] = '\0';
      break;

    /* Name used for displaying */
    case DAV_PROPID_displayname:
      if (resource->collection && strcmp(resource->info->sfn, "/") != 0) {
        /* Copy only the last part, stripping the last '/' */
        p2 = resource->info->sfn + strlen(resource->info->sfn);
        for(p = p2 - 2;
            p != resource->info->sfn && *p != '/';
            --p);
        ++p;
        i = p2 - p;
        memcpy(value, p, i);
        value[i] = '\0';
      }
      else {
        p = strrchr(resource->info->sfn, '/');
        if (p)
          strncpy(value, p + 1, sizeof(buffer));
      }
      break;

    /* E-Tag, used for caching */
    case DAV_PROPID_getetag:
      strncpy(value, resource->hooks->getetag(resource), sizeof(buffer));
      break;

    /* Content-type (MIME) */
    case DAV_PROPID_getcontenttype:
      if (resource->collection)
        return DAV_PROP_INSERT_NOTDEF;
      else
        if (dav_dpm_mime_get(buffer, sizeof(buffer), strrchr(resource->info->sfn, '/') + 1) == NULL)
          return DAV_PROP_INSERT_NOTDEF;
      break;

    /* Grid Unique IDentifier */
    case DAV_PROPID_guid:
      if (resource->info->gstat.atime == 0)
        Cns_statr(resource->info->sfn, &(resource->info->gstat));
      strncpy(value, resource->info->gstat.guid, sizeof(buffer));
      break;

    /* Mode bits */
    case DAV_PROPID_mode:
      snprintf(buffer, sizeof(buffer), "0%o", resource->info->stat.filemode);
      break;

    /* The owner user. Name, or UID if not found */
    case DAV_PROPID_owner:
      snprintf(buffer, sizeof(buffer), "%u", resource->info->stat.uid);
      break;

    /* The group. Name, or GID if not found */
    case DAV_PROPID_group:
      snprintf(buffer, sizeof(buffer), "%u", resource->info->stat.gid);
      break;

    /* The checksum type */
    case DAV_PROPID_sumtype:
      if (resource->info->gstat.atime == 0)
        Cns_statr(resource->info->sfn, &(resource->info->gstat));
      strncpy(value, resource->info->gstat.csumtype, sizeof(buffer));
      break;

    /* The checksum value */
    case DAV_PROPID_sumvalue:
      if (resource->info->gstat.atime == 0)
        Cns_statr(resource->info->sfn, &(resource->info->gstat));
      strncpy(value, resource->info->gstat.csumvalue, sizeof(buffer));
      break;

    /* Associated comment */
    case DAV_PROPID_comment:
      value[0] = '\0';
      if (Cns_getcomment(resource->info->sfn, value) < 0) {
        if (serrno != ENOENT) {
          ap_log_rerror(APLOG_MARK, APLOG_ERR, serrno,
                        resource->info->request,
                        "Could not get the comment for %s", resource->info->sfn);
        }
        else {
          /* We are cool with ENOENT, as there may be no comment */
          serrno = 0;
        }
      }
      break;

    /* List of replicas */
    case DAV_PROPID_replicas:
      if (resource->collection)
        return DAV_PROP_INSERT_NOTDEF;
      else if (Cns_getreplica(resource->info->sfn, NULL, NULL, &nreplies, &replicas) == 0 && nreplies) {
        /* In this case, the stack allocated data probably is not enough */
        value = dav_ns_serialize_replicas(resource->pool, nreplies, replicas,
                                          resource->info->s_conf->type == DAV_DPM_NODE_LFC);
        free(replicas);
      }
      else {
        value[0] = '\0';
      }
      break;

    /* File ID */
    case DAV_PROPID_fileid:
      snprintf(value, sizeof(buffer), "%llu", resource->info->stat.fileid);
      break;

    /* File class */
    case DAV_PROPID_fileclass:
      snprintf(value, sizeof(buffer), "%d", resource->info->stat.fileclass);
      break;

    /* Status */
    case DAV_PROPID_status:
      value[0] = resource->info->stat.status;
      value[1] = '\0';
      break;

    default:
      /* Why are we here? Where did that come from? */
      ap_log_rerror(APLOG_MARK, APLOG_ERR, 0,
                    resource->info->request, "Something weird happened here: propid=%d", propid);
      return DAV_PROP_INSERT_NOTDEF;
  }

  /* Insert into the output */
  global_ns = dav_get_liveprop_info(propid, &dav_dpm_liveprop_group, &prop_info);
  
  /* Asking for the value? */
  if (what == DAV_PROP_INSERT_VALUE) {
    s = apr_psprintf(resource->pool, "<lp%d:%s>%s</lp%d:%s>",
                     global_ns, prop_info->name, value, global_ns, prop_info->name);
  }
  /* Asking for the name? */
  else if (what == DAV_PROP_INSERT_NAME) {
    s = apr_psprintf(resource->pool, "<lp%d:%s/>",
                     global_ns, prop_info->name);
  }
  /* Or just asking if it supported? */
  else {
    /* DAV_PROP_INSERT_SUPPORTED */
    s = apr_psprintf(resource->pool,
                    "<D:supported-live-property D:name=\"%s\" "
                    "D:namespace=\"%s\"/>",
                    prop_info->name, dav_dpm_namespace_uris[prop_info->ns]);
  }
  apr_text_append(resource->pool, phdr, s);

  /* Done */
  return what;
}

/**
 * Is the property writable?
 * @param resource The resource being queried
 * @param propid   The property ID
 * @return         1 if it is, 0 if not
 */
static int dav_dpm_is_writable(const dav_resource *resource, int propid)
{
  const dav_liveprop_spec *prop_info;

  dav_get_liveprop_info(propid, &dav_dpm_liveprop_group, &prop_info);
  return prop_info->is_writable;
}

/**
 * Check if the passed value for the given property is good enough
 * @param resource  The resource being modified
 * @param elem      The value for the property
 * @param operation Set or delete
 * @param context   Can be used to pass values to the rest of the functions
 * @param defer_to_dead Let it be to be managed by dead properties DB
 * @return          NULL on success
 */
static dav_error *dav_dpm_patch_validate(const dav_resource *resource,
                                const apr_xml_elem *elem,
                                int operation,
                                void **context,
                                int *defer_to_dead)
{
  dav_elem_private         *priv = elem->priv;
  dav_dpm_liveprop_context *value;
  const apr_text           *data;
  int                       i;
  struct tm                 st_time;
  dav_error                *err;

  /* Removing these properties is not supported */
  if (operation != DAV_PROP_OP_SET)
    return dav_dpm_new_error(resource->pool, EINVAL, HTTP_METHOD_NOT_ALLOWED,
                             "Live properties can not be removed (%s:%s)",
                             dav_dpm_namespace_uris[elem->ns], elem->name);

  /* Get pointer to the value */
  data = elem->first_cdata.first;
  if (!data)
    return dav_dpm_new_error(resource->pool, EINVAL, HTTP_CONFLICT,
                             "No value specified for the property");

  /* Create context */
  value = apr_pcalloc(resource->pool, sizeof(dav_dpm_liveprop_context));
  *context = value;

  /* Check the value is right */
  switch(priv->propid) {
    case DAV_PROPID_getlastmodified: case DAV_PROPID_lastaccess:
      if (strptime(data->text, "%a, %d %b %Y %H:%M:%S", &st_time) == NULL)
        return dav_dpm_new_error(resource->pool, EINVAL, HTTP_CONFLICT,
                                "Invalid date and time format");
      value->tim = timegm(&st_time);
      break;
      
    case DAV_PROPID_executable: case DAV_PROPID_executable_apache:
      /* We want a single letter: T or F */
      value->flag = data->text[0];
      if (strlen(data->text) != 1 || (value->flag != 'T' && value->flag != 'F'))
        return dav_dpm_new_error(resource->pool, EINVAL, HTTP_CONFLICT,
                                 "The 'executable' property expects only 'T' or 'F'");
      break;

    case DAV_PROPID_mode:
      /* Only a number */
      value->mode = strtol(data->text, NULL, 0);
      if (value->mode == 0 && data->text[0] != '0')
        return dav_dpm_new_error(resource->pool, EINVAL, HTTP_CONFLICT,
                                 "A numeric string must be provided. Tailing non-digit will be ignored");
      break;

    case DAV_PROPID_owner:
      for (i = 0; isdigit(data->text[i]); ++i);

      if(data->text[i] == '\0') {
        value->uid = atoi(data->text);
      }
      else {
        if (Cns_getusrbynam((char*)data->text, &value->uid) < 0)
          return dav_dpm_new_error(resource->pool, serrno, 0,
                                   "Could not map the user %s", data->text);
      }
      if (resource->info->user->uid != 0 && !has_group(resource->info->user, 0))
        return dav_dpm_new_error(resource->pool, EPERM, 0,
                                 "Only root or an user belonging to root group can change the owner");
      break;
      
    case DAV_PROPID_group:
      for (i = 0; isdigit(data->text[i]); ++i);

      if(data->text[i] == '\0') {
        value->gid = atoi(data->text);
      }
      else {
        if (Cns_getgrpbynam((char*)data->text, &value->gid) < 0)
          return dav_dpm_new_error(resource->pool, serrno, 0,
                                   "Could not map the group %s", data->text);
      }
      if (resource->info->user->uid != 0 && !has_group(resource->info->user, 0) &&
          resource->info->user->uid != resource->info->stat.uid)
        return dav_dpm_new_error(resource->pool, EPERM, 0,
                                 "Only the owner, root or an user belonging to root group can change the group");
      else if (!has_group(resource->info->user, value->gid))
        return dav_dpm_new_error(resource->pool, EPERM, 0,
                                 "The user doesn't belong to the new group");
      break;

    case DAV_PROPID_comment:
      /* The value is not empty, so we are good */
      value->string = (char*)data->text;
      break;

    case DAV_PROPID_replicas:
      err = dav_ns_deserialize_replicas(resource->pool, data->text,
                                        &value->replica_array);
      if (err != NULL)
        return err;

      /* Validate action and values */
      for (i = 0; i < value->replica_array.nreplicas; ++i) {
        struct Cns_filereplicax *replica = &value->replica_array.replicas[i];

        switch (value->replica_array.action[i]) {
          case '+': case '-': case 'M':
            break;
          default:
            return dav_dpm_new_error(resource->pool, EINVAL,
                                     HTTP_BAD_REQUEST,
                                     "Action '%c' for replica not supported",
                                     value->replica_array.action[i]);
        }

        if (replica->sfn[0] == '\0')
          return dav_dpm_new_error(resource->pool, EINVAL,
                                   HTTP_BAD_REQUEST,
                                   "The attribute 'replica' is empty or not defined");
      }

      break;

    default:
      /* Unsupported */
      *defer_to_dead = 0;
      return dav_dpm_new_error(resource->pool, EINVAL, 0, "Unknown live property and dead properties not supported");
  }

  return NULL;
}

/**
 * Finally, set the property
 * @param resource  The resource being modified
 * @param elem      The value of the property
 * @param operation Set or remove
 * @param context   As set by dav_dpm_patch_validate
 * @param rollback_ctx In case the change has to be undone, put here relevant information
 * @return
 */
static dav_error *dav_dpm_patch_exec(const dav_resource *resource,
                                     const apr_xml_elem *elem,
                                     int operation,
                                     void *context,
                                     dav_liveprop_rollback **rollback_ctx)
{
  dav_elem_private         *priv  = elem->priv;
  dav_dpm_liveprop_context *value = context;
  struct Cns_filestat      stat;
  struct Cns_fileid        uniqueid;
  int                      n;
  struct utimbuf           timebuf;

  /* We already have the proper values in *value (*context) */
  switch(priv->propid) {
    case DAV_PROPID_getlastmodified:
      timebuf.modtime = value->tim;
      timebuf.actime  = resource->info->stat.atime;
      if (Cns_utime(resource->info->sfn, &timebuf) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not change mtime of %s", resource->info->sfn);
      break;

    case DAV_PROPID_lastaccess:
      timebuf.modtime = resource->info->stat.mtime;
      timebuf.actime  = value->tim;
      if (Cns_utime(resource->info->sfn, &timebuf) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not change atime of %s", resource->info->sfn);
      break;
      
    case DAV_PROPID_executable: case DAV_PROPID_executable_apache:
      /* Get the mode, set the executable flag and let the next case do the job */
      if (Cns_stat(resource->info->sfn, &stat) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not stat %s", resource->info->sfn);
      if (value->flag == 'T')
        value->mode = stat.filemode | S_IXUSR;
      else if(stat.filemode & S_IXUSR)
        value->mode = stat.filemode ^ S_IXUSR;
      else
        value->mode = stat.filemode;

    case DAV_PROPID_mode:
      if (Cns_chmod(resource->info->sfn, value->mode) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not change the mode");
      break;

    case DAV_PROPID_owner:
      if (Cns_chown(resource->info->sfn, value->uid, -1) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not change the uid");
      break;

    case DAV_PROPID_group:
      if (Cns_chown(resource->info->sfn, -1, value->gid) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not change the gid");
      break;

    case DAV_PROPID_comment:
      if (Cns_setcomment(resource->info->sfn, value->string) < 0)
        return dav_dpm_new_error(resource->pool, serrno, 0,
                                 "Could not set the comment");
      break;

    case DAV_PROPID_replicas:
      for (n = 0; n < value->replica_array.nreplicas; ++n) {
        struct physical_loc      location;
        struct Cns_filereplicax* replica = &value->replica_array.replicas[n];
        
        uniqueid.fileid = resource->info->stat.fileid;
        strncpy(uniqueid.server, resource->info->s_conf->ns_host, CA_MAXHOSTNAMELEN);

        /* Need the host */
        dav_dpm_split_turl(resource->pool, replica->sfn, &location);

        switch (value->replica_array.action[n]) {
          case '+':
              if (Cns_addreplica(NULL, &uniqueid, location.host,
                                 replica->sfn, replica->status, replica->f_type,
                                 replica->poolname, replica->fs) != 0)
              return dav_dpm_new_error(resource->pool, serrno, 0,
                                       "Could not add the replica %s",
                                       replica->sfn);
            break;
          case '-':
            if (Cns_delreplica(NULL, &uniqueid, replica->sfn) != 0)
              return dav_dpm_new_error(resource->pool, serrno, 0,
                                      "Could not delete the replica %s",
                                      replica->sfn);
            break;
          default:
            if (replica->ltime != 0)
              Cns_setrltime(replica->sfn, replica->ltime);
            if (replica->atime != 0)
              Cns_setratime(replica->sfn);
            if (replica->f_type != 0)
              Cns_setrtype(replica->sfn, replica->f_type);
            if (replica->status != 0)
              Cns_setrstatus(replica->sfn, replica->status);
            break;
        }
      }
      break;

    default:
      /* We will never enter here due to previous checking */
      break;
  }

  return NULL;
}

/**
 * Nothing to do here, really
 * @param resource
 * @param operation
 * @param context
 * @param rollback_ctx
 */
static void dav_dpm_patch_commit(const dav_resource *resource,
                                 int operation,
                                 void *context,
                                 dav_liveprop_rollback *rollback_ctx)
{
  return;
}

/**
 * Undo a change when a failure occurs
 * @param resource  The resource being modified
 * @param operation The operation performed (set, delete...)
 * @param context   As set by dav_dpm_patch_validate
 * @param rollback_ctx The rollback context
 * @return          NULL on success
 */
static dav_error *dav_dpm_patch_rollback(const dav_resource *resource,
                                         int operation,
                                         void *context,
                                         dav_liveprop_rollback *rollback_ctx)
{
  /* If setting the property failed, it wasn't set at all, so there is
   * no need to rollback */
  return NULL;
}

/* Properties handlers */
static const dav_hooks_liveprop dav_hooks_liveprop_dpm =
{
    dav_dpm_insert_prop,
    dav_dpm_is_writable,
    dav_dpm_namespace_uris,
    dav_dpm_patch_validate,
    dav_dpm_patch_exec,
    dav_dpm_patch_commit,
    dav_dpm_patch_rollback
};

/* Properties */
const dav_liveprop_group dav_dpm_liveprop_group =
{
    dav_dpm_props,          /* Properties specification */
    dav_dpm_namespace_uris, /* List of namespaces       */
    &dav_hooks_liveprop_dpm /* Handlers                 */
};

/**
 * Find the property ID so the value can be retrieved later
 * @param resource The resource being processed
 * @param ns_uri   The property namespace (e.g. DAV:)
 * @param name     The property name      (e.g. getcontentlength)
 * @param hooks    Where to put the liveprop hooks
 * @return         The property ID, or 0 if not found
 */
int dav_dpm_find_liveprop(const dav_resource *resource, const char *ns_uri,
                           const char *name, const dav_hooks_liveprop **hooks)
{
  const dav_liveprop_spec *scan;
  int ns;

  /* One of us, not one of them */
  if (resource->hooks != &dav_hooks_repository_dpm)
    return 0;

  /* Locate the namespace in the namespace table */
  for (ns = 0; dav_dpm_namespace_uris[ns] != NULL; ++ns) {
    if (strcmp(ns_uri, dav_dpm_namespace_uris[ns]) == 0)
      break;
  }

  /* No one matched? */
  if (dav_dpm_namespace_uris[ns] == NULL)
    return 0;

  /* Now look for the property ID */
  for (scan = dav_dpm_props; scan->name != NULL; ++scan) {
    if (ns == scan->ns && strcmp(name, scan->name) == 0) {
      *hooks = &dav_hooks_liveprop_dpm;
      return scan->propid;
    }
  }

  /* Property name not found */
  return 0;
}

/**
 * Insert all supported liveprops into the output
 * @param r        The request
 * @param resource The resource being processed
 * @param what     What to insert
 * @param phdr     Where to insert
 */
void dav_dpm_insert_all_liveprops(request_rec *r, const dav_resource *resource,
                                  dav_prop_insert what, apr_text_header *phdr)
{
  const dav_liveprop_spec *scan;
  
  /* One of us, not one of them */
  if (resource->hooks != &dav_hooks_repository_dpm)
    return;

  /* Non-existing have no properties */
  if (!resource->exists)
    return;

  /* Insert all liveprops */
  for (scan = dav_dpm_props; scan->name != NULL; ++scan)
    dav_dpm_insert_prop(resource, scan->propid, what, phdr);

  return;
}
