#ifndef MOD_DAV_DPM_H_
#define MOD_DAV_DPM_H_

#include "../mod_dav/mod_dav.h"
#include <apr_shm.h>
#include <apr_hash.h>
#include <Cns_api.h>
#include <httpd.h>
#include <http_config.h>
#include <openssl/sha.h>
#include <openssl/ssl.h>

#define FLAG_HANDLE_GET   1
#define DPM_MAX_ERROR_LEN 512
#define MAX_FQANS         32

#define DAV_TOKEN "DAV-DPM interface request"

#define DAV_DPM_RFC2068 0
#define DAV_DPM_RFC3339 1

#define DAV_DPM_WRITE             0x01
#define DAV_DPM_DISK_DIR_ALLOW    0x02
#define DAV_DPM_DISK_DIR_REDIRECT 0x04
#define DAV_DPM_REMOTE_COPY       0x08

#if  SHA_DIGEST_LENGTH * 2 > CA_MAXDPMTOKENLEN
#define DAV_DPM_MAXTOKENLEN SHA_DIGEST_LENGTH * 2
#else
#define DAV_DPM_MAXTOKENLEN CA_MAXDPMTOKENLEN
#endif

enum {
  WAITING,
  FAILED,
  SUCCESS
};

/** Prototype of a function that generates a token from a request
 * @param r       The request
 * @param token   Where to put the token
 * @param st      The size of the buffer where the token will be put
 * @param sfn     The logical file name
 * @param path    The physical path
 * @param dn      The user DN
 * @param r_token The DPM request token
 * @param extra   Whatever was passed as argument in the configuration (i.e. password)
 * @return NULL on success
 */
typedef dav_error* (*generate_token_f)(request_rec *r, char *token, size_t st,
                                       const char *sfn, const char *path, const char *dn,
                                       const char *r_token, const char *extra);

/** Prototype of a function that validates a token
 * @param r       The request
 * @param token   Where to put the token
 * @param sfn     The logical file name
 * @param path    The physical path
 * @param dn      The user DN
 * @param extra   Whatever was passed as argument in the configuration (i.e. password)
 * @return NULL on success
 */
typedef dav_error* (*validate_token_f)(request_rec *r, const char *token,
                                       const char *sfn, const char *path, const char *dn,
                                       const char *extra);

/* Node types */
typedef enum {
  DAV_DPM_NODE_DISK,
  DAV_DPM_NODE_HEAD,
  DAV_DPM_NODE_LFC
} node_type;

typedef unsigned char dir_flags;

/* Data that needs to be available outside */
extern module AP_MODULE_DECLARE_DATA dav_dpm_module;
extern const dav_hooks_repository    dav_hooks_repository_dpm;
extern const dav_liveprop_group      dav_dpm_liveprop_group;
extern const dav_hooks_db            dav_hooks_db_dbm;

/*
 * User mapping cache register
 */
struct dav_dpm_user_map_cache_entry {
  uid_t uid;
  gid_t gids[MAX_FQANS];
  char  session_id[SSL_MAX_SSL_SESSION_ID_LENGTH];
  long  expiration;
};
typedef struct dav_dpm_user_map_cache_entry dav_dpm_user_map_cache_entry;

/* Data structure for server configuration */
struct dav_dpm_server_conf
{
    const char *dpm_host, *ns_host, *dav_host;
    const char *csec_mech;
    node_type  type;
    
    apr_shm_t *session_cache_shm;
    unsigned   session_cache_size;
};
typedef struct dav_dpm_server_conf dav_dpm_server_conf;
/* Data structure for directory configuration */
struct dav_dpm_dir_conf
{
    const char *base, *anon_user, *proxy_cache;
    dir_flags   flags;

    apr_hash_t *filesystems;

    generate_token_f  generate_token;
    validate_token_f  validate_token;
    const char       *token_extra;
};
typedef struct dav_dpm_dir_conf dav_dpm_dir_conf;

/* Security information about the user */
struct dav_dpm_user_info {
  char  *fqans[MAX_FQANS];
  char  *principal, *voname;
  int    nfqans;
  uid_t  uid;
  gid_t  gids[MAX_FQANS];
};
typedef struct dav_dpm_user_info dav_dpm_user_info;

/*
 * This structure handles the data the module needs to keep track of
 * a resource coming from a DPM service
 * typedef done in mod_dav.h
 */
struct physical_loc {
  char     protocol[CA_MAXPROTOLEN];
  char     host    [CA_MAXHOSTNAMELEN];
  char     path    [CA_MAXPATHLEN];
  unsigned port;

  char token       [DAV_DPM_MAXTOKENLEN];
};
struct dav_resource_private {
  char                 sfn[CA_MAXSFNLEN];
  char                *redirect; /* Allocated only when needed */
  const char          *root_dir; /* Point to the root_dir passed by get_resource */
  dav_dpm_user_info   *user;
  struct Cns_filestat  stat;
  struct Cns_filestatg gstat;
  struct physical_loc  location;

  request_rec    *request;
  apr_table_t    *query;

  dav_dpm_server_conf *s_conf;
  dav_dpm_dir_conf    *d_conf;
};

/**
 * Insert all supported liveprops into the output
 * @param r        The request
 * @param resource The resource being processed
 * @param what     What to insert
 * @param phdr     Where to insert
 */
void dav_dpm_insert_all_liveprops(request_rec *r, const dav_resource *resource,
                                  dav_prop_insert what, apr_text_header *phdr);

/**
 * Find the property ID so the value can be retrieved later
 * @param resource The resource being processed
 * @param ns_uri   The property namespace (e.g. DAV:)
 * @param name     The property name      (e.g. getcontentlength)
 * @param hooks    Where to put the liveprop hooks
 * @return         The property ID, or 0 if not found
 */
int dav_dpm_find_liveprop(const dav_resource *resource, const char *ns_uri,
                           const char *name, const dav_hooks_liveprop **hooks);

/**
 * Interacts with mod_gridsite through environment variables and initializes
 * an instance of dav_dpm_user_info
 * Everything is created using Apache pool, so don't bother freeing
 * Returns NULL on error, and sets errno
 * @param r      The request
 * @param s_conf The server configuration
 * @param d_conf The directory configuration
 * @return  A pointer to an initialized dav_dpm_user_info point
 */
dav_dpm_user_info *dav_dpm_get_user_info(request_rec *r, dav_dpm_server_conf *s_conf, dav_dpm_dir_conf *d_conf);

/**
 *  Create a dav error with proper error message and HTTP status
 * @param p         The pool used to allocate memory
 * @param err_no    The error code (usually, the value of errno or serrno)
 * @param http_code If specified, this will be used instead of mapping err_no
 * @param format    Additional string to print (format)
 * @return          A initialized dav_error instance. Must not be freed.
 */
DAV_DECLARE(dav_error*) dav_dpm_new_error(apr_pool_t *p, int err_no, int http_code, const char *format, ...);

/* UTILS */

/**
 * Return a string with the date-time formated
 * @param buffer Where to put the formated date
 * @param maxlen The size of the buffer
 * @param tstamp The timestamp
 * @param format Format flag
 * @return       The length of the printed string
 */
int dav_dpm_format_datetime(char *buffer, size_t maxlen, time_t tstamp, unsigned format);

/**
 * Add a response to the walk resource
 * @param wres  The walk resource
 * @param error The error to add
 */
void dav_dpm_add_response(dav_walk_resource *wres, dav_error *error);

/**
 * Parses a query (everything after a ?) and initializes a table with it
 * @param pool
 * @param query
 * @return
 */
apr_table_t *dav_dpm_parse_query(apr_pool_t *pool, const char *query);

/**
 * Gets the location from a TURL. HTTP/S aware. It will asume HTTPS for
 * unrecognised protocols.
 * @param pool     The pool
 * @param turl     The TURL to split
 * @param location Where to put the parsed URL
 */
void dav_dpm_split_turl(apr_pool_t *pool, const char *turl, struct physical_loc *location);

/**
 * Generates a printable version of the filemode
 * @param str  Where to put the string
 * @param mode The mode
 * @return     str
 */
char *dav_dpm_mode_str(char *str, mode_t mode);

/**
 * Generates a printable version of the size
 * @param str  Where to put the string
 * @param max  Max length
 * @param size The size in bytes
 * @return     str
 */
char *dav_dpm_size_str(char *str, size_t max, size_t size);

/**
 * Initializes the internal structures with the mapping
 * @param pool     The pool
 * @param map_path The path to the mime file
 */
void dav_dpm_mime_init(apr_pool_t *pool, const char *map_path);

/**
 * Returns the mime type associated with the filename's extension
 * If no extension is found, or not correspondance, NULL will be returned
 * @param buffer   Where to put the value
 * @param maxlen   The size of the buffer
 * @param filename The file name
 * @return buffer if found, NULL if not
 */
const char *dav_dpm_mime_get(char *buffer, size_t maxlen, const char *filename);

/**
 * Performs a remote copy
 * @param src    The original resource
 * @param remote The remote URL
 * @param depth  The copy depth
 * @return       NULL on success
 */
dav_error *dav_dpm_remote_copy(const dav_resource *src, const char *remote,
                               int depth);

/**
 * Get the physical location of the file
 * @param info Input/output parameter. sfn must be set. location will be set.
 * @param pool Pool for memory allocation
 * @return NULL on success
 */
dav_error *dav_dpm_get_physical_location(dav_resource_private *info,
                                         apr_pool_t *pool);

/* Current token implementation */
dav_error *dpm_generate_token(request_rec *r, char *token, size_t st,
                              const char *sfn, const char *path, const char *dn,
                              const char *r_token, const char *extra);
dav_error *dpm_validate_token(request_rec *r, const char *token,
                              const char *sfn, const char *path, const char *dn,
                              const char *extra);
dav_error *passwd_generate_token(request_rec *r, char *token, size_t st,
                                 const char *sfn, const char *path, const char *dn,
                                 const char *r_token, const char *extra);
dav_error *passwd_validate_token(request_rec *r, const char *token,
                               const char *sfn, const char *path, const char *dn,
                               const char *extra);
dav_error *file_generate_token(request_rec *r, char *token, size_t st,
                               const char *sfn, const char *path, const char *dn,
                               const char *r_token, const char *extra);
dav_error *file_validate_token(request_rec *r, const char *token,
                               const char *sfn, const char *path, const char *dn,
                               const char *extra);

/**
 * Returns 1 is SSL is being used
 */
char ssl_used(conn_rec *connection);

#endif
