#include "mod_dav_dpm.h"
#include <apr_file_io.h>
#include <apr_strings.h>
#include <Cns_api.h>
#include <dpm_api.h>
#include <http_core.h>
#include <http_log.h>
#include <http_protocol.h>
#include <http_request.h>
#include <serrno.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

/* Used for internal recursivity */
#define DAV_WALKTYPE_POSTFIX 0x8000
#define DAV_CALLTYPE_POSTFIX 1000

/* Advanced declarations */
static dav_error *dav_dpm_walk(const dav_walk_params *params, int depth,
                               dav_response **response);

/* Structure used for the walking */
struct dav_dpm_walker_context {
  const dav_walk_params *params;

  dav_walk_resource walk_resource;
 
  dav_buffer    sfn;
  dav_buffer    uri;
};
typedef struct dav_dpm_walker_context dav_dpm_walker_context;

/* Internal structure for streams (typedef by mod_dav.h) */
struct dav_stream {
  const dav_resource *resource;
  apr_file_t         *fd;
  unsigned            has_range;
  apr_size_t          written;
};

/**
 * Get the physical location of the file
 * @param info Input/output parameter. sfn must be set. location will be set.
 * @param pool Pool for memory allocation
 * @return NULL on success
 */
dav_error *dav_dpm_get_physical_location(dav_resource_private *info,
                                         apr_pool_t *pool)
{
  int                       nreplies;
  const char               *err_str   = NULL;
  dav_error                *token_err = NULL;

  if (info->s_conf->dpm_host)
  {
    /* DPMHost is configured, so look for the disk */
    char                     *proto, *u_token;
    char                      r_token[CA_MAXDPMTOKENLEN + 1];
    int                       status;
    struct dpm_getfilereq     request;
    struct dpm_getfilestatus *statuses;

    request.from_surl  = (char*)info->sfn;
    /* Figure out what do these mean, and where should they come from */
    request.lifetime   = 0;
    request.f_type     = '\0';
    request.s_token[0] = '\0';
    request.ret_policy = '\0';
    request.flags      = 0;

    proto   = apr_pstrdup(pool, "rfio");
    u_token = apr_pstrdup(pool, DAV_TOKEN);

    /* Request and keep waiting until we get something */
    if(dpm_get(1, &request, 1, &proto, u_token, 0, r_token, &nreplies, &statuses) < 0)
      return dav_dpm_new_error(pool, serrno, 0, "Can not get %s", info->sfn);
    else if (!nreplies)
      return dav_dpm_new_error(pool, EINVAL, 0, "No replies for %s", info->sfn);

    switch (statuses[0].status) {
      case DPM_QUEUED: case DPM_RUNNING: case DPM_ACTIVE:
        status = WAITING;
        break;
      case DPM_READY: case DPM_SUCCESS: case DPM_DONE:
        status = SUCCESS;
        break;
      default:
        status = FAILED;
    }

    /* Wait while not ready */
    while (status == WAITING) {
      if (dpm_getstatus_getreq(r_token, 1, &request.from_surl, &nreplies, &statuses) < 0) {
        if (serrno == 0)
          serrno = statuses[0].status & 0xFFF;
        status  = FAILED;
        err_str = apr_pstrcat(pool, "Could not get replicas: ", statuses[0].errstring, NULL);
      }
      else if (!nreplies) {
        if (serrno == 0)
          serrno = statuses[0].status & 0xFFF;
        status  = FAILED;
        err_str = apr_pstrcat(pool, "Error querying ", info->sfn, NULL);
      }
      else {
        switch(statuses[0].status) {
          case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
            if (serrno == 0)
              serrno = statuses[0].status & 0xFFF;
            status  = FAILED;
            err_str = apr_pstrcat(pool, "Could not get replicas: ",
                                  statuses[0].errstring, NULL);
            break;
          case DPM_READY: case DPM_SUCCESS: case DPM_DONE:
            status = SUCCESS;
            break;
        }
      }
    }

    /* Get TURL */
    if (status == SUCCESS && statuses[0].turl) {
      dav_dpm_split_turl(pool, statuses[0].turl, &info->location);
      /* Free */
      dpm_free_gfilest(nreplies, statuses);
      /* Get token */
      token_err = info->d_conf->generate_token(info->request,
                                               info->location.token, DAV_DPM_MAXTOKENLEN,
                                               info->sfn, info->location.path,
                                               info->user->principal,
                                               r_token,
                                               info->d_conf->token_extra);
    }
    else if (status == SUCCESS) {
      status  = FAILED;
      serrno  = EINVAL;
      err_str = "TURL empty";
    }

    /* Return error, if any */
    if (status == FAILED && serrno != ENOENT) {
      return dav_dpm_new_error(pool, serrno, 0, err_str);
    }
    else if (serrno == ENOENT) {
      return dav_dpm_new_error(pool, ENOENT, 0,
                               "No replicas for %s", info->sfn);
    }
    else if (token_err) {
      return token_err;
    }

    /* Is it me? */
    if (apr_strnatcasecmp(info->location.host, info->request->server->server_hostname) != 0) {
      const char *query = info->request->parsed_uri.query;

      info->redirect = apr_psprintf(pool, "%s://%s:%u%s?token=%s&path=%s%s%s",
                                   info->location.protocol,
                                   info->location.host,
                                   info->location.port,
                                   info->request->parsed_uri.path,
                                   info->location.token,
                                   info->location.path,
                                   query ? "&"   : "",
                                   query ? query : "");
    }
    else {
      info->redirect = NULL;
    }
  }
  else {
    /* No DPMHost, so use an alternative: get replicas and redirect */
    struct Cns_filereplica *replicas;
    struct physical_loc     location;
    
    if (Cns_getreplica(info->sfn, NULL, NULL, &nreplies, &replicas) != 0)
      return dav_dpm_new_error(pool, serrno, 0,
                               "Error getting replicas for %s", info->sfn);
    if (nreplies < 1)
      return dav_dpm_new_error(pool, serrno, 0,
                               "No replicas for %s", info->sfn);

    dav_dpm_split_turl(pool, replicas[0].sfn, &location);

    info->redirect = apr_psprintf(pool, "%s://%s:%u%s",
                                 location.protocol,
                                 location.host, location.port,
                                 location.path);
    
    ap_log_rerror(APLOG_MARK, 0, APLOG_INFO, info->request,
                  "%s located at %s", info->sfn, info->redirect);

    free(replicas);
  }

  /* Last check: if we are not redirecting, make sure it is allowed to
   * access the file locally */
  if (!info->redirect) {
    apr_hash_index_t *iter;
    unsigned          allowed;
    const char       *key, *value, *location;
    apr_ssize_t       key_size;

    /* Path may start with double /, so jump over it */
    for (location = info->location.path;
         *(location+1) == '/';
         ++location);

    /* Search for a match */
    for (iter = apr_hash_first(pool, info->d_conf->filesystems), allowed = 0;
         iter && !allowed;
         iter = apr_hash_next(iter)) {
      apr_hash_this(iter, (const void**)&key, &key_size, (void*)&value);
      allowed = (strncmp(key, location, key_size) == 0);
    }

    if (!allowed) {
      return dav_dpm_new_error(pool, EACCES, 0,
                               "%s not under an allowed path",
                               location);
    }
  }

  return NULL;
}

/**
 * Return the parent path of "path"
 * @param pool Where to allocate space
 * @param path The path
 * @return     The parent path
 */
static char *dav_dpm_dirname(apr_pool_t *pool, const char *path)
{
  char *parent;
  int   i;

  parent = apr_pstrcat(pool, path, NULL);

  /* We skip the last character so if path is a directory, we are fine */
  for(i = strlen(parent) - 2; i >= 0; --i) {
    if (parent[i] == '/') {
      parent[i + 1] = '\0'; /* Keep the slash */
      break;
    }
  }

  return parent;
}

/**
 * This function is used internally to create new resources
 * @param r          The request to associate
 * @param root_dir   The root dir in Apache, so redirects can be build properly
 * @param sfn        The SFN to link to the resource (must NOT change in the future)
 * @param uri        The URI (without the query!) (must NOT change in the future)
 * @param user       The associated user. If NULL, will be queried and initialized
 * @param resource   Where to put it
 * @return           NULL on success
 */
static dav_error *dav_dpm_internal_get_resource(request_rec *r,
                                                const char *root_dir,
                                                const char *sfn,
                                                const char *uri,
                                                dav_dpm_user_info *user,
                                                dav_resource **resource)
{
  dav_resource_private *info;
  int                   exists;
  const char           *p;

  /* Initialize info */
  info   = apr_pcalloc(r->pool, sizeof(dav_resource_private));

  info->request  = r;
  info->root_dir = root_dir;
  strncpy(info->sfn, sfn, CA_MAXSFNLEN);

  info->s_conf = ap_get_module_config(r->server->module_config,  &dav_dpm_module);
  info->d_conf = ap_get_module_config(r->per_dir_config,         &dav_dpm_module);

  /* User info */
  if (user) {
    info->user = user;
  }
  else {
    if (!(info->user = dav_dpm_get_user_info(r, info->s_conf, info->d_conf)))
      return dav_dpm_new_error(r->pool, serrno, 0, "Can not authenticate the user");

    /* If we were already here for this connection, no need to set again */
    if (apr_table_get(r->connection->notes, "dav_dpm_user_set_authz") == NULL) {
      Cns_client_setAuthorizationId(info->user->uid, info->user->gids[0], "GSI", info->user->principal);
      dpm_client_setAuthorizationId(info->user->uid, info->user->gids[0], "GSI", info->user->principal);
      
      apr_table_setn(r->connection->notes, "dav_dpm_user_set_authz", (void*)1);
    }

    if (info->user->nfqans) {
      Cns_client_setVOMS_data(info->user->voname, info->user->fqans, info->user->nfqans);
      dpm_client_setVOMS_data(info->user->voname , info->user->fqans, info->user->nfqans);
    }
  }

  /* Stat */
  if (Cns_stat(info->sfn, &(info->stat)) != 0 && serrno != ENOENT)
    return dav_dpm_new_error(r->pool, serrno, 0, "Can not stat %s", info->sfn);

  /* NULL resources are used before creating something */
  exists = (serrno != ENOENT);

  /* Get query and token, if available */
  info->query = dav_dpm_parse_query(r->pool, r->parsed_uri.query);
  p = apr_table_get(info->query, "token");
  if (p)
    strncpy(info->location.token, p, DAV_DPM_MAXTOKENLEN);
  
  /* If there is a token, we want the physical location */
  if (info->location.token[0]) {
    char                     *sfns[] = { apr_pstrdup(r->pool, info->sfn) };
    int                       nreplies;

    if (r->method_number == M_PUT) {
      /* For PUT methods, the token is the DPM put token */
      struct dpm_putfilestatus *statuses;

      if (dpm_getstatus_putreq((char*)info->location.token, 1, sfns, &nreplies, &statuses) < 0) {
        return dav_dpm_new_error(r->pool, serrno, HTTP_BAD_REQUEST,
                                 "Could not retrieve the destination for %s",
                                 info->location.token);
      }

      /* Since we allow multistreamed, be careful the putreq is not finished */
      if (statuses[0].status != DPM_READY && statuses[0].status != DPM_SUCCESS) {
        return dav_dpm_new_error(r->pool, EACCES, 0,
                                 "File upload was already finished. Can not modify it! (%x)",
                                 statuses[0].status);
      }

      /* Set the destination */
      dav_dpm_split_turl(r->pool, statuses[0].turl, &info->location);
      dpm_free_pfilest(nreplies, statuses);
      
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "PUT with token (%s:%s (%s)",
                    info->location.host, info->location.path, info->location.token);

      /* For all effects, act like this does not exist (NULL resource) */
      exists = 0;
    }
    else {
      /* For others, the token is configurable */
      const char *path;
      dav_error  *err;

      path = apr_table_get(info->query, "path");
      if (path) {
        strncpy(info->location.path, path, CA_MAXPATHLEN);
      } else {
        int                    nreplies;
        struct dpm_filestatus *statuses;
        char                  *surls[] = { (char*)info->sfn };

        /* Can be a "close" HEAD */
        if (dpm_putdone((char*)info->location.token, 1, surls, &nreplies, &statuses) < 0) {
          if (serrno != EINVAL)
            return dav_dpm_new_error(r->pool, serrno, 0, "Error on dpm_putdone for %s", info->sfn);
        }
        else {
          /* Succeeded, so recurse to stat the file */
          ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                         "putdone! Recursive call to get the final state");
          r->parsed_uri.query = NULL;
          return dav_dpm_internal_get_resource(r, root_dir, sfn, uri, user, resource);
        }
        
        /* It wasn't */
        return dav_dpm_new_error(r->pool, EINVAL, HTTP_BAD_REQUEST,
                                 "A path is needed in the request");
      }

      /* Validate the path */
      err = info->d_conf->validate_token(r, info->location.token,
                                         info->sfn, info->location.path,
                                         info->user->principal,
                                         info->d_conf->token_extra);
      if (err) {
        return err;
      }

      strncpy(info->location.host, info->request->server->server_hostname, CA_MAXHOSTNAMELEN);

      /* Beware! Protocol not initialized */
      if (ssl_used(r->connection))
        strcpy(info->location.protocol, "https");
      else
        strcpy(info->location.protocol, "http");
      info->location.port = r->server->port;


      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "GET? with token (%s:%s (%s)",
                    info->location.host, info->location.path, info->location.token);
    }
  } /* if (info->location.token[0]) */

  /* Create and return resource */
  *resource = apr_pcalloc(r->pool, sizeof(dav_resource));
  (*resource)->type   = DAV_RESOURCE_TYPE_REGULAR;
  (*resource)->exists = exists;
  (*resource)->collection = ((info->stat.filemode & S_IFDIR) != 0);
  (*resource)->uri    = uri;
  (*resource)->info   = info;
  (*resource)->hooks  = &dav_hooks_repository_dpm;
  (*resource)->pool   = r->pool;

  if (exists)
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Resource for %s (%s) found (dir=%d)",
                  (*resource)->uri, (*resource)->info->sfn, (*resource)->collection);
  else
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "NULL resource for %s (%s) created",
                  (*resource)->uri, (*resource)->info->sfn);

  return NULL;
}

/**
 * Generate the resource used for later operations
 * First security checkings can be done here (like user mapping)
 * @param r                The request coming from Apache
 * @param root_dir         The root dir of the request (as configured in Apache with Location)
 * @param label            Used with version control. Not relevant.
 * @param use_checked_in   Used with version control. Not relevant.
 * @param resource         Where to put the generated resource
 * @return                 NULL on success. A dav_error instance if not.
 */
static dav_error *dav_dpm_get_resource(request_rec *r,
                                       const char *root_dir,
                                       const char *label,
                                       int use_checked_in,
                                       dav_resource **resource)
{
  dav_dpm_server_conf  *s_conf;
  dav_dpm_dir_conf     *d_conf;
  char                 *sfn, *path;
  dav_error            *err;
  int                   len;

  s_conf = ap_get_module_config(r->server->module_config, &dav_dpm_module);
  d_conf = ap_get_module_config(r->per_dir_config,        &dav_dpm_module);

  /* Compose path */
  len  = strlen(root_dir);
  if (root_dir[len - 1] == '/') {
    char *tmp = apr_pstrdup(r->pool, root_dir);
    --len;
    tmp[len] = '\0';
    root_dir = tmp;
  }

  path = r->parsed_uri.path ? r->parsed_uri.path : "";
  sfn = apr_pstrcat(r->pool,
                    d_conf->base, path + len,
                    NULL);

  /* Create resource */
  err = dav_dpm_internal_get_resource(r, root_dir, sfn, path, NULL, resource);
  if (err)
    return err;

  /* Directories MUST finish with '/', so redirect if not */
  len = strlen(path);
  if ((*resource)->collection && (len == 0 || path[len - 1] != '/')) {
    (*resource)->info->redirect = apr_pstrcat(r->pool,
                                              ap_escape_uri(r->pool, path),
                                              "/",
                                              r->args ? "?" : "",
                                              r->args ? r->args : "",
                                              NULL);
    (*resource)->uri = apr_pstrcat(r->pool, (*resource)->uri, "/", NULL);
    strncat((*resource)->info->sfn, "/", CA_MAXSFNLEN);
  }

  /* If Dir and Disk: redirect, deny, serve? */
  if ((*resource)->collection && s_conf->type == DAV_DPM_NODE_DISK) {
    if (d_conf->flags & DAV_DPM_DISK_DIR_REDIRECT) {
      /* If / was appended, no need to build the path again */
      const char *rest = (*resource)->info->redirect;
      if (!rest) {
        rest = apr_pstrcat(r->pool,
                           ap_escape_uri(r->pool, path),
                           r->args ? "?" : "",
                           r->args ? r->args : "",
                           NULL);
      }
      /* Final URL */
      (*resource)->info->redirect = apr_pstrcat(r->pool, "https://",
                                                s_conf->dav_host,
                                                rest,
                                                NULL);
      /* This case is special. Do the redirect right here, right now */
      apr_table_set(r->headers_out, "Location", (*resource)->info->redirect);
      return dav_new_error(r->pool, HTTP_MOVED_PERMANENTLY, 0, (*resource)->info->redirect);
    }
    else if (!(d_conf->flags & DAV_DPM_DISK_DIR_ALLOW)) {
      return dav_dpm_new_error(r->pool, EACCES, 0,
                               "The disk node can not serve directories");
    }
  }

  /* We are done here */
  return NULL;
}

/**
 * Get the parent of a resource (if any)
 * @param resource        The resource to process
 * @param parent_resource The parent, if any
 * @return                NULL on success
 */
static dav_error *dav_dpm_get_parent(const dav_resource *resource,
                                     dav_resource **parent_resource)
{
  char        *sfn, *uri;
  request_rec *dup_rec;

  /* The parent of the root is itself (kinda) */
  if (strcmp(resource->info->sfn, "/") == 0) {
    *parent_resource = (dav_resource*)resource;
    return NULL;
  }

  /* Strip last element of the path */
  sfn = dav_dpm_dirname(resource->pool, resource->info->sfn);
  uri = dav_dpm_dirname(resource->pool, resource->uri);

  /* Duplicate request, removing query from the parsed URI.
   * This is because on PUT requests with query string, it will remain
   * in the parent, making get_resource to behave like it does not exist :-(
   */
  dup_rec  = apr_pcalloc(resource->pool, sizeof(request_rec));
  *dup_rec = *resource->info->request;
  dup_rec->parsed_uri.query = NULL;

  /* Return whatever the internal says */
  return dav_dpm_internal_get_resource(dup_rec,
                                       resource->info->root_dir,
                                       sfn, uri, resource->info->user,
                                       parent_resource);
}

/**
 * Compare two resources to check if they are the same
 * @param res1 Resource 1
 * @param res2 Resource 2
 * @return     1 if they are the same
 */
static int dav_dpm_is_same(const dav_resource *res1, const dav_resource *res2)
{
  const char *destination;
  apr_uri_t  uri1, uri2;

  /* Parse resource 1 */
  apr_uri_parse(res1->pool, res1->uri, &uri1);

  /* Parse resource 2 */
  destination = apr_table_get(res2->info->request->headers_in, "destination");
  if (destination)
    apr_uri_parse(res2->pool, destination, &uri2);
  else
    apr_uri_parse(res2->pool, res2->uri, &uri2);

  /* Host may not appear */
  if (!uri1.hostname)
    uri1.hostname = res1->info->request->server->server_hostname;
  if (!uri2.hostname)
    uri2.hostname = res2->info->request->server->server_hostname;

  /* Compare */
  return strcmp(uri1.hostname, uri2.hostname) == 0 && strcmp(uri1.path, uri2.path) == 0;
}

/**
 * Check if a given resource is the parent of another one
 * @param res1 Parent to check
 * @param res2 Child
 * @return     1 if res1 is the parent of res2
 */
static int dav_dpm_is_parent(const dav_resource *res1, const dav_resource *res2)
{
  char *res2_parent_sfn;

  res2_parent_sfn = dav_dpm_dirname(res2->pool, res2->info->sfn);

  return strcmp(res1->info->sfn, res2_parent_sfn) == 0;
}

/**
 * Open a stream. Used for PUT
 * @param resource The resource to open
 * @param mode     The mode (basically, trucate or random, which we don't support)
 * @param stream   Where to put the created structure
 * @return         NULL on success
 */
static dav_error *dav_dpm_open_stream(const dav_resource *resource,
                                      dav_stream_mode mode,
                                      dav_stream **stream)
{
  dav_resource_private     *info;
  unsigned                  has_range;
  const char               *range, *length;

  info = resource->info;
  
  /* Must be writable (if we are here, they want to write) */
  if (!(resource->info->d_conf->flags & DAV_DPM_WRITE))
    return dav_dpm_new_error(resource->pool, EACCES, 0,
                             "Configured as read-only endpoint (%s)",
                             resource->uri);

  /* Must not exist */
  if (resource->exists)
    return dav_dpm_new_error(resource->pool, EACCES, HTTP_METHOD_NOT_ALLOWED,
                             "The file already exists (%s)",
                             resource->info->sfn);

  /* If content-length is 0, call create instead of put! */
  length = apr_table_get(info->request->headers_in, "content-length");
  if (length != NULL) {
    size_t clen = atol(length);
    if (clen == 0) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    info->request,
                    "PUT with content-length 0. Creating ns entry.");

      if (Cns_creat(info->sfn, 0644) != 0)
        return dav_dpm_new_error(resource->pool, errno, 0,
                                "Could not create empty file %s", info->sfn);

      *stream = (dav_stream*)calloc(1, sizeof(dav_stream));

      return NULL;
    }
  }

  /* NS alone don't support PUTs */
  if (!resource->info->s_conf->dpm_host)
    return dav_dpm_new_error(resource->pool, ENOSYS, 0,
                             "Name servers do not support PUTs");

  /*  Range header? */
  range = apr_table_get(resource->info->request->headers_in, "content-range");
  has_range = (range != NULL);
  if (has_range) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                  resource->info->request,
                  "Range: %s", range);
  }

  /* If token not set, call dpm_put and see where do we have to go */
  if (!info->location.token[0]) {
    struct dpm_putfilereq     reqfile;
    struct dpm_putfilestatus *statuses;
    char                     *proto, *u_token;
    char                     *err_str = NULL;
    char                     *space_token;
    char                      r_token[CA_MAXDPMTOKENLEN + 1];
    int                       status, nreplies, errcode;

    /* Build request */
    proto   = apr_pstrdup(resource->pool, "rfio");
    u_token = (char*)DAV_TOKEN;

    reqfile.to_surl        = apr_pstrdup(resource->pool, info->sfn);
    reqfile.f_type         = 'P';
    reqfile.lifetime       = 0;
    reqfile.requested_size = 0;
    reqfile.f_lifetime     = 0;
    reqfile.ret_policy     = '\0';
    reqfile.ac_latency     = '\0';

    /* If there is a space token, use it */
    space_token = (char*)apr_table_get(info->query, "spacetoken");
    if (space_token != NULL) {
      char **space_ids;

      if (dpm_getspacetoken(space_token, &nreplies, &space_ids) < 0)
        return dav_dpm_new_error(resource->pool, serrno, HTTP_BAD_REQUEST, "Can not get the space token %s", space_token);

      strncpy(reqfile.s_token, space_ids[0], sizeof(reqfile.s_token));
      while (nreplies >= 1) {
        free(space_ids[nreplies - 1]);
        nreplies--;
      }
      free(space_ids);

      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, resource->info->request,
                    "Token with ID %s will be used", reqfile.s_token);
    }
    else {

      reqfile.s_token[0] = '\0';
    }

    /* Do the PUT */
    if (dpm_put(1, &reqfile, 1, &proto, u_token, 0, 0, r_token, &nreplies, &statuses) < 0)
      return dav_dpm_new_error(resource->pool, serrno, 0, "Can not put %s", info->sfn);
    else if (nreplies < 0)
      return dav_dpm_new_error(resource->pool, EINVAL, 0, "No replies for %s", info->sfn);

    /* Wait for completion */
    switch (statuses[0].status) {
      case DPM_QUEUED: case DPM_RUNNING: case DPM_ACTIVE:
        status = WAITING;
        break;
      case DPM_READY: case DPM_SUCCESS:
        status = SUCCESS;
        break;
      default:
        status = FAILED;
    }
    err_str = NULL;

    while (status == WAITING) {
      if (dpm_getstatus_putreq(r_token, 1, &reqfile.to_surl, &nreplies, &statuses) < 0) {
        if (serrno == 0)
          serrno = statuses[0].status & 0xFFF;
        err_str = apr_pstrdup(resource->pool, statuses[0].errstring);
        status  = FAILED;
      }
      else if (!nreplies) {
        if (serrno == 0)
          serrno = EINVAL;
        err_str = apr_pstrcat(resource->pool, "Error querying for put ", info->sfn, NULL);
        status  = FAILED;
      }
      else {
        switch(statuses[0].status) {
          case DPM_FAILED: case DPM_ABORTED: case DPM_EXPIRED:
            if (serrno == 0)
              serrno = statuses[0].status & 0xFFF;
            err_str = apr_pstrcat(resource->pool,
                                  "Could not get a destination: ",
                                  statuses[0].errstring, NULL);
            status = FAILED;
            break;
          case DPM_READY: case DPM_SUCCESS:
            status = SUCCESS;
            break;
        }
      }
    }

    /* Get the TURL if available */
    if (status == SUCCESS && statuses[0].turl != NULL) {
      dav_dpm_split_turl(resource->pool, statuses[0].turl, &info->location);
      strncpy(info->location.token, r_token, DAV_DPM_MAXTOKENLEN);
    }
    else if (status == SUCCESS) {
      serrno  = EINVAL;
      err_str = "TURL empty";
      status  = FAILED;
    }

    /* Free the return list and fail if something happened */
    errcode = serrno;
    dpm_free_pfilest(nreplies, statuses);
    if (status == FAILED)
      return dav_dpm_new_error(resource->pool, errcode, 0, err_str);
  
    /* If the destination is not the current host OR
     * there is a Range header, redirect (on multistreamed, the token will be needed) */
    if (apr_strnatcasecmp(info->location.host, info->request->server->server_hostname) != 0
        || has_range) {
      info->redirect = apr_psprintf(resource->pool, "%s://%s:%u%s?token=%s%s%s",
                                   info->location.protocol,
                                   info->location.host,
                                   info->location.port,
                                   info->request->parsed_uri.path,
                                   r_token,
                                   info->request->args ? "&" : "",
                                   info->request->args ? info->request->args : "");
    }
    else {
      info->redirect = NULL;
    }

    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                  info->request,
                  "PUT request to be done in %s:%s",
                  info->location.host, info->location.path);

    /* Redirect */
    if(info->redirect) {
      apr_table_set(info->request->headers_out,
                    "Location", info->redirect);
      /* Standard says 301, but clients will retry in that case with a GET.
       * Curl specifies anything > 303 means redoing the whole thing */
      return dav_new_error(resource->pool, HTTP_TEMPORARY_REDIRECT, 0,
                           info->redirect);
    }
  } /* if (!info->location.token[0])  */

  /* This is the destination, so create an internal descriptor
   * and there we go */
  *stream = apr_pcalloc(resource->pool, sizeof(dav_stream));
  (*stream)->resource  = resource;
  (*stream)->has_range = has_range;
  (*stream)->written   = 0;
    
  if (apr_file_open(&(*stream)->fd, resource->info->location.path,
                    APR_WRITE | APR_CREATE | APR_BUFFERED,
                    APR_OS_DEFAULT, resource->pool) != APR_SUCCESS) {
    return dav_dpm_new_error(resource->pool, errno, 0,
                             "Could not open %s",
                             resource->info->location.path);
  }
  return NULL;
}

/**
 * Close a stream
 * @param stream The stram to close
 * @param commit Commit the change
 * @return       NULL on success
 */
static dav_error *dav_dpm_close_stream(dav_stream *stream, int commit)
{
  dav_resource_private *info;

  /* Empty NS entry! */
  if (stream->fd == NULL) {
    return NULL;
  }

  /* Close */
  apr_file_close(stream->fd);
  info = stream->resource->info;

  if(commit && !stream->has_range) {
    /* Commit the change */
    char                  *surls[] = { (char*)info->sfn };
    int                    nreplies;
    struct dpm_filestatus *statuses;

    if (dpm_putdone((char*)info->location.token, 1, surls, &nreplies, &statuses) < 0) {
      return dav_dpm_new_error(stream->resource->pool, serrno, 0,
                               "dpm_putdone error (%s)", info->sfn);
    }

    dpm_free_filest(nreplies, statuses);
    /* That is! */
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                  info->request,
                  "dpm_putdone on %s (%zu written)",
                  info->sfn, stream->written);
  }
  else if(!commit) {
    /* Remove the file (best effort, it will be remove lately if failed) */
    apr_file_remove(info->location.path, stream->resource->pool);
  }
  else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                  info->request,
                  "Commit with range, so do not finish yet (%zu written)",
                  stream->written);
  }

  return NULL;
}

/**
 * Write to a stream
 * @param stream  The stream where to write
 * @param buf     The data to write
 * @param bufsize How much data there is in the buffer
 * @return        NULL on success
 */
static dav_error *dav_dpm_write_stream(dav_stream *stream, const void *buf,
                                       apr_size_t bufsize)
{
  /* Empty NS entry! */
  if (stream->fd == NULL) {
    return NULL;
  }

  /* Write */
  if (apr_file_write_full(stream->fd, buf, bufsize, NULL) != APR_SUCCESS)
    return dav_dpm_new_error(stream->resource->pool, errno, 0,
                             "Error writing to %s",
                             stream->resource->info->location.path);
  stream->written += bufsize;
  return NULL;
}

/**
 * Moves to an specific position within a stream
 * @param stream       The stream to move
 * @param abs_position Where to move
 * @return             NULL on success
 */
static dav_error *dav_dpm_seek_stream(dav_stream *stream, apr_off_t abs_position)
{
  /* Empty NS entry! */
  if (stream->fd == NULL) {
    return NULL;
  }

  /* Seek */
  if (apr_file_seek(stream->fd, APR_SET, &abs_position) != APR_SUCCESS)
    return dav_dpm_new_error(stream->resource->pool, errno, 0,
                            "Error seeking %s (%lu)",
                            stream->resource->info->location.path, abs_position);

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                stream->resource->info->request,
                "Seek %s (%lu)",
                stream->resource->info->location.path, (unsigned long)abs_position);

  return NULL;
}

/**
 * Sets the HTTP headers, like the size of the content, the modification date...
 * @param r        The request
 * @param resource The resource generated previously
 * @return         NULL on success
 */
static dav_error *dav_dpm_set_headers(request_rec *r, const dav_resource *resource)
{
  dav_resource_private *info;
  dav_error            *err;
  const char           *etag, *match;

  /* Null resource */
  if(!resource->exists)
    return NULL;

  info = resource->info;

  /* We can not delay this more
   * File? Then, get replica and redirect, or continue and serve ourselves */
  if (!resource->collection) {
    /* If the token is available, we already got this */
    if(!info->location.token[0]) {
      if ((err = dav_dpm_get_physical_location(info, r->pool)))
        return err;
    }
  }

  /* Redirect */
  if (info->redirect) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Resource %s located at %s",
                   info->sfn, info->redirect);

    apr_table_setn(r->headers_out, "Location", info->redirect);
    return dav_new_error(r->pool, HTTP_MOVED_TEMPORARILY, 0, info->redirect);
  }

  if (resource->collection) {
    /* text/html for directories */
    ap_set_content_type(r, "text/html");
  }
  else {
    char buffer[256];
    ap_set_content_length(r, info->stat.filesize);
    /* Mime-type */
    if (dav_dpm_mime_get(buffer, sizeof(buffer), strrchr(resource->info->sfn, '/') + 1))
      ap_set_content_type(r, buffer);
    /* Content-Location when there is no redirection, but might be useful */
    apr_table_setn(r->headers_out, "Content-Location",
                   apr_psprintf(resource->pool, "%s://%s:%u%s?token=%s&path=%s",
                                info->location.protocol,
                                info->location.host,
                                info->location.port,
                                info->request->parsed_uri.path,
                                info->location.token,
                                info->location.path));
  }

  /* Nice thing is, Apache takes care of implementing this */
  apr_table_setn(r->headers_out, "Accept-Ranges", "bytes");

  /* ETag */
  etag = resource->hooks->getetag(resource);
  apr_table_setn(r->headers_out, "ETag", etag);

  /* If "If-None-Match" is present, and match the ETag, cut here to save time */
  match = apr_table_get(r->headers_in, "If-None-Match");
  if (match && strcmp(match, etag) == 0)
    return NULL;

  /* DPNS saves the time in seconds since epoch, while Apache expects microseconds */
  ap_update_mtime(r, info->stat.mtime * 1000000);
  ap_set_last_modified(r);
  
  return NULL;
}

/**
 * This is the second most important function, as it delivers the content of the
 * file (or the directory if not DAV)
 * @param resource The resource generated previously
 * @param output   Here is where the output must be written
 * @return         NULL on success
 */
static dav_error *dav_dpm_deliver(const dav_resource *resource, ap_filter_t *output)
{
  apr_bucket_brigade   *bb;
  apr_bucket           *bkt;
  dav_resource_private *info;

  info = resource->info;
  bb   = apr_brigade_create(resource->pool, output->c->bucket_alloc);

  if (resource->collection) {
    /* Directories */
    Cns_DIR              *dir;
    struct Cns_direnstat *entry;
    char                 *up;

    /* Use sessions to speedup */
    Cns_startsess((char*)info->s_conf->ns_host, "dav_dpm_deliver directory");

    /* Open */
    if (!(dir = Cns_opendir(info->sfn)))
      return dav_dpm_new_error(resource->pool, serrno, 0,
                               "Could not open directory %s",
                               info->sfn);

    /* Generate HTML content for the dir */
    if (strcmp(resource->uri, "/") != 0)
      up = "[<a href=\"..\">&crarr;</a>]";
    else
      up = "";

    ap_fprintf(info->request->output_filters, bb,
               "<!DOCTYPE html PUBLIC \"-//W3C//DTD XHTML 1.0 Strict//EN\" \"http://www.w3.org/TR/xhtml1/DTD/xhtml1-strict.dtd\">\n"
               "<html xmlns=\"http://www.w3.org/1999/xhtml\">\n"
               "<head>\n"
               "<meta http-equiv=\"content-type\" content=\"text/html;charset=utf-8\"/>\n"
               "<link rel=\"stylesheet\" type=\"text/css\" href=\"/static/css/lcgdm-dav.css\"/>\n"
               "<title>%s</title>\n"
               "</head>\n"
               "<body>\n"
               "<div id=\"header\"><h1>%s%s</h1></div>\n",
                resource->uri, resource->uri, up);

    ap_fputs(info->request->output_filters, bb,
             "<table id=\"ft\">\n"
             "<thead><tr>\n"
             "<th class=\"ft-mode\">Mode</th>"
             "<th class=\"ft-uid\">UID</th>"
             "<th class=\"ft-gid\">GID</th>"
             "<th class=\"ft-size\">Size</th>"
             "<th class=\"ft-datetime\">Modified</th>"
             "<th class=\"ft-name\">Name</th>"
             "</tr></thead>\n");

    while((entry = Cns_readdirx(dir))) {
      char *tail, *class, mode_buf[11], size_buf[32], date_buf[64];

      /* Directories end with / */
      if (S_ISDIR(entry->filemode)) {
        tail  = "/";
        class = "ft-folder";
      }
      else if (S_ISLNK(entry->filemode)) {
        tail  = "";
        class = "ft-link";
      }
      else {
        tail  = "";
        class = "ft-file";
      }

      /* Human-readable */
      dav_dpm_mode_str(mode_buf, entry->filemode);
      if (entry->filemode & S_IFDIR)
        snprintf(size_buf, sizeof(size_buf), "%d", entry->nlink);
      else
        dav_dpm_size_str(size_buf, sizeof(size_buf), entry->filesize);

      dav_dpm_format_datetime(date_buf, sizeof(date_buf), entry->mtime, DAV_DPM_RFC2068);

      /* Print */
      ap_fprintf(info->request->output_filters, bb,
                 "<tr>"
                 "<td>%s</td>"
                 "<td>%d</td><td>%d</td>"
                 "<td>%s</td>"
                 "<td>%s</td>"
                 "<td class=\"%s\"><a href=\"%s%s\">%s%s</a></td>"
                 "</tr>\n",
                 mode_buf,
                 entry->uid, entry->gid,
                 size_buf,
                 date_buf,
                 class, entry->d_name, tail, entry->d_name, tail);
    }

    if(serrno)
      return dav_dpm_new_error(resource->pool, serrno, 0,
                               "Error retrieving the entries for %s",
                               info->sfn);

    ap_fputs(info->request->output_filters, bb, "</table>\n");
    ap_fputs(info->request->output_filters, bb, "<div id=\"footer\">\n");

    if(info->user->principal)
      ap_fprintf(info->request->output_filters, bb,
                 "<p><span id=\"requestby\">Request by %s (%s)</span>\n",
                 info->user->principal,
              info->user->voname ? info->user->voname : "No proxy");

    ap_fputs(info->request->output_filters, bb, "<br/>Powered by LCGDM-DAV</p>\n");
    ap_fputs(info->request->output_filters, bb, "</div></body>\n</html>");

    if(Cns_closedir(dir))
      return dav_dpm_new_error(resource->pool, serrno, 0,
                               "Could not close %s",
                               info->sfn);

    /* End session */
    Cns_endsess();

  }
  else {
    /* File */
    apr_file_t *fd;

    if (!info->location.path[0])
      return dav_dpm_new_error(resource->pool, ENOENT, 0,
                               "No replicas for %s", info->sfn);

    if (apr_file_open(&fd, info->location.path, APR_READ | APR_SENDFILE_ENABLED, 0, resource->pool) != APR_SUCCESS)
      return dav_dpm_new_error(resource->pool, errno, 0,
                               "Could not open %s",
                               info->location.path);

    bkt = apr_brigade_insert_file(bb, fd, 0, info->stat.filesize, resource->pool);
  }

  /* Close and flush the output */
  bkt = apr_bucket_eos_create(output->c->bucket_alloc);
  APR_BRIGADE_INSERT_TAIL(bb, bkt);
  if (ap_pass_brigade(info->request->output_filters, bb) != APR_SUCCESS)
    return dav_dpm_new_error(resource->pool, EIO, 0, "Could not write EOS to filter.");

  /* All OK */
  return NULL;
}

/**
 * Create a new collection (i.e. directory)
 * @param resource The resource to be created
 * @return         NULL on success
 */
static dav_error *dav_dpm_create_collection(dav_resource *resource)
{
  /* Must be writable */
  if (!(resource->info->d_conf->flags & DAV_DPM_WRITE))
    return dav_dpm_new_error(resource->pool, EACCES, 0,
                             "Configured as read-only endpoint (%s)",
                             resource->uri);

  /* If a Disk, listing must be allowed */
  if (resource->info->s_conf->type == DAV_DPM_NODE_DISK &&
      !(resource->info->d_conf->flags & DAV_DPM_DISK_DIR_ALLOW))
    return dav_dpm_new_error(resource->pool, EACCES, 0,
                             "Creating directories in disks not allowed");

  /* Must not exist */
  if (resource->exists)
    return dav_dpm_new_error(resource->pool, EACCES, 0,
                             "The file already exists (%s)",
                             resource->info->sfn);

  /* Create */
  if (Cns_mkdir(resource->info->sfn, 0755) != 0)
    return dav_dpm_new_error(resource->pool, serrno,
                             (serrno == ENOENT) ? HTTP_CONFLICT : 0,
                             "Could not create the directory %s",
                             resource->info->sfn);

  resource->exists     = 1;
  resource->collection = 1;
  if (Cns_stat(resource->info->sfn, &(resource->info->stat)) != 0) {
    /* It has just been created, we shouldn't enter here, but just in case */
    return dav_dpm_new_error(resource->pool, serrno, 0,
                             "Cns_stat failed just after the creation of %s",
                              resource->info->sfn);
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                resource->info->request,
               "%s created", resource->info->sfn);

  return NULL;
}

/**
 * Copy a resource to an non-existing one
 * Only allowed for remote copies
 * @param src       The source. It must exist
 * @param dst       The destination. Must not exist
 * @param depth     Copy in depth? (Collections)
 * @param response  The response to the copy process
 * @return          NULL on success
 */
static dav_error *dav_dpm_copy(const dav_resource *src, dav_resource *dst,
                               int depth, dav_response **response)
{
  *response = NULL;
  return dav_dpm_remote_copy(src,
                             apr_table_get(src->info->request->headers_in, "destination"),
                             depth);
}

/**
 * Move a resource
 * @param src       The source. It must exist
 * @param dst       The destination. It must not exist
 * @param response  The response to the move process
 * @return          NULL on success
 */
static dav_error *dav_dpm_move(dav_resource *src, dav_resource *dst,
                               dav_response **response)
{
  /* Must be writable */
  if (!(src->info->d_conf->flags & DAV_DPM_WRITE))
    return dav_dpm_new_error(dst->pool, EACCES, 0,
                             "Configured as read-only endpoint (%s)",
                             dst->uri);

  /* Change the name, that's all */
  if (Cns_rename (src->info->sfn, dst->info->sfn) < 0) {
    return dav_dpm_new_error(dst->pool, serrno, 0,
                             "Could not move from %s to %s",
                             src->info->sfn, dst->info->sfn);
  }

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                src->info->request,
                "Moved %s to %s", src->info->sfn, dst->info->sfn);

  *response = NULL;
  return NULL;
}

/**
 * Remove a resource from DPM, either file or directory
 * NOT recursive, resource has to be a leaf, or this will fail
 * @param resource The resource to remove
 * @return         NULL on success
 */
static dav_error *dav_dpm_internal_remove(const dav_resource *resource)
{
  int errnum = 0;

  if (resource->collection) {
    /* Remove a directory is straigh-forward */
    if (Cns_rmdir(resource->info->sfn) == 0) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    resource->info->request,
                    "Directory %s removed", resource->info->sfn);
    }
    else {
      errnum = serrno;
    }
  }
  else {
    /* A file is not much more difficult :-) */
    int                    nreplies;
    char                  *sfns[] = {apr_pstrdup(resource->pool, resource->info->sfn)};
    struct dpm_filestatus *statuses;

    if (dpm_rm(1, sfns, &nreplies, &statuses) != 0)
      errnum = serrno;
    dpm_free_filest(nreplies, statuses);

    if (errnum == 0) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
                    resource->info->request,
                    "File %s removed", resource->info->sfn);
    }
  }

  if (errnum) {
    ap_log_rerror(APLOG_MARK, APLOG_ERR, 0,
                  resource->info->request,
                  "Could not remove %s", resource->info->sfn);
    
    return dav_dpm_new_error(resource->pool, errnum, 0,
                             "Could not remove %s", resource->info->sfn);
  }
  
  return NULL;
}

/**
 * Callback function for remove walking
 * @param wres      The walking resource
 * @param calltype  Calltype
 * @return          NULL on success
 */
static dav_error *dav_dpm_remove_callback(dav_walk_resource *wres, int calltype)
{
  /* Only remove elements that exist, and empty directories */
  if (wres->resource->exists &&
      (!wres->resource->collection || calltype == DAV_CALLTYPE_POSTFIX)) {

    dav_error *err;

    err = dav_dpm_internal_remove(wres->resource);

    /* We add to the multistatus, not as return */
    if (err)
      dav_dpm_add_response(wres, err);    
  }
  return NULL;
}

/**
 * Remove a resource
 * @param resource The resource to remove
 * @param response The response to the request
 * @return         NULL on success
 */
static dav_error *dav_dpm_remove(dav_resource *resource,
                                 dav_response **response)
{
  dav_error *err;

  /* Be careful: mod_dav will try to remove the file when copying, if source
   * and dest happen to have the same path (even if the host is different.
   * Just ignore those requests.
   */
  if (resource->info->request->method_number == M_COPY) {
    ap_log_rerror(APLOG_MARK, APLOG_INFO, 0,
                  resource->info->request,
                  "Ignoring remove requests on COPY");
    return NULL;
  }

  /* Must be writable */
  if (!(resource->info->d_conf->flags & DAV_DPM_WRITE))
    return dav_dpm_new_error(resource->pool, EACCES, 0,
                             "Configured as read-only endpoint (%s)",
                             resource->uri);

  
  if (resource->collection) {
    /* Recursive if collection */
    dav_walk_params  params = { 0 };
    dav_response    *multi_status;

    params.walk_type = DAV_WALKTYPE_POSTFIX;
    params.func      = dav_dpm_remove_callback;
    params.pool      = resource->pool;
    params.root      = resource;

    if ((err = dav_dpm_walk(&params, DAV_INFINITY, &multi_status)) != NULL)
      return err;

    if ((*response = multi_status) != NULL) {
      /* Multistatus */
      return dav_new_error(resource->pool, HTTP_MULTI_STATUS, 0,
                           "Errors on multistatus");
    }

    resource->exists = 0;
    resource->collection = 0;

    return NULL;
  }
  else {
    /* File is pretty easy */
    err = dav_dpm_internal_remove(resource);
    if (err)
      return err;
    resource->exists     = 0;
    resource->collection = 0;
    return NULL;
  }
}

/**
 * Recursive function to walk through the filesystem
 * @param ctx The context
 * @param depth The maximum depth
 * @return NULL on success
 */
static dav_error *dav_dpm_walker(dav_dpm_walker_context *ctx, int depth)
{
  dav_error            *err = NULL;
  Cns_DIR              *dir;
  struct Cns_direnstat *entry;
  apr_pool_t           *pool;
  const dav_resource   *this_res;
  dav_resource          reusable;
  apr_pool_t           *subpool;

  pool = ctx->params->pool;

  /* Save for later */
  this_res = ctx->walk_resource.resource;

  /* Call the function with the current root */
  err = (*ctx->params->func)(&(ctx->walk_resource),
                               ctx->walk_resource.resource->collection ?
                                 DAV_CALLTYPE_COLLECTION: DAV_CALLTYPE_MEMBER);
  if (err)
    return err;
  if (depth == 0 || !ctx->walk_resource.resource->collection)
    return NULL;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0,
              ctx->walk_resource.resource->info->request,
              "Descending into %s", ctx->sfn.buf);

  /* Open the directory */
  if ((dir = Cns_opendir(ctx->sfn.buf)) == NULL) {
    dav_dpm_add_response(&ctx->walk_resource,
                         dav_dpm_new_error(pool, serrno, 0,
                         "Could not open the directory %s",
                         ctx->sfn.buf));
    return NULL;
  }

  /* Create the subpool */
  apr_pool_create(&subpool, ctx->walk_resource.pool);

  /* Initialize the reusable resource */
  reusable = *this_res;
  reusable.info = NULL;
  reusable.uri  = NULL;
  reusable.info = apr_pcalloc(reusable.pool, sizeof(dav_resource_private));
  reusable.info->request = this_res->info->request;
  reusable.info->user    = this_res->info->user;
  reusable.info->s_conf  = this_res->info->s_conf;
  reusable.info->d_conf  = this_res->info->d_conf;
  reusable.pool = subpool;

  ctx->walk_resource.resource = &reusable;

  /* Scan the directory */
  while((entry = Cns_readdirx(dir))) {
    apr_size_t    uri_prev_len;
    apr_size_t    path_prev_len;
    int len;

    apr_pool_clear(subpool);

    /* Append the file to path and URI */
    len = strlen(entry->d_name);
    dav_buffer_place_mem(pool, &ctx->uri, entry->d_name, len + 1, 1);
    dav_buffer_place_mem(pool, &ctx->sfn, entry->d_name, len + 1, 1);

    strncpy(reusable.info->sfn, ctx->sfn.buf, CA_MAXPATHLEN);
    
    /* dav_buffer_place_mem do NOT adjust len */
    uri_prev_len  = ctx->uri.cur_len;
    path_prev_len = ctx->sfn.cur_len;
    ctx->sfn.cur_len  += len;
    ctx->uri.cur_len  += len;

    if (S_ISDIR(entry->filemode)) {
      ctx->sfn.cur_len++;
      ctx->uri.cur_len++;

      ctx->uri.buf[ctx->uri.cur_len - 1] = ctx->sfn.buf[ctx->sfn.cur_len - 1] = '/';
      ctx->uri.buf[ctx->uri.cur_len]     = ctx->sfn.buf[ctx->sfn.cur_len]     = '\0';
    }

    /* Set "entry-related" fields */
    reusable.uri        = ctx->uri.buf;
    reusable.collection = S_ISDIR(entry->filemode);

    /* Care with this. Cns_filestat is a subset of Cns_dirstat,
     * with same order and types. If this correspondency is broken, this will
     * fail badly */
    reusable.info->stat = *((struct Cns_filestat*)entry);

     /* Attach or descend */
    if (!ctx->walk_resource.resource->collection) {
      /* Add file */
      err = ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_MEMBER);
      if (err)
        break;  
    }
    else {
      /* Recursive into child */
      if ((err = dav_dpm_walker(ctx, depth - 1)) != NULL)
        break;
    }

    /* Restore length, so next dav_buffer_place_mem overrites current entry */
    ctx->uri.cur_len  = uri_prev_len;
    ctx->sfn.cur_len  = path_prev_len;
  }

  apr_pool_destroy(subpool);

  /* Release resources */
  if (Cns_closedir(dir))
    return dav_dpm_new_error(pool, serrno, 0,
                             "Could not close the directory %s", ctx->sfn.buf);

  if (err)
    return err;

  /* Recover */
  ctx->walk_resource.resource    = this_res;
  ctx->uri.buf[ctx->uri.cur_len] = '\0';
  ctx->sfn.buf[ctx->sfn.cur_len] = '\0';

  /* Postfix iteration */
  if (ctx->params->walk_type & DAV_WALKTYPE_POSTFIX) {
    return ctx->params->func(&(ctx->walk_resource), DAV_CALLTYPE_POSTFIX);
  }

  /* There we go */
  return NULL;
}

/**
 * Walk through the file system
 * @param params    Walk parameters
 * @param depth     In depth?
 * @param response  The response
 * @return          NULL on success
 */
static dav_error *dav_dpm_walk(const dav_walk_params *params, int depth,
                               dav_response **response)
{
  dav_dpm_walker_context  ctx = {0};
  dav_error              *err;

  ctx.params = params;

  ctx.walk_resource.walk_ctx = params->walk_ctx;
  ctx.walk_resource.pool     = params->pool;
  ctx.walk_resource.resource = params->root;

  dav_buffer_init(params->pool, &ctx.sfn, params->root->info->sfn);
  dav_buffer_init(params->pool, &ctx.uri, params->root->uri);

  Cns_startsess((char*)params->root->info->s_conf->ns_host, "dav_dpm_walk directory");
  err = dav_dpm_walker(&ctx, depth);
  Cns_endsess();
  
  *response = ctx.walk_resource.response;

  return err;
}

/**
 * Get an ETag
 * @param resource The resource from where retrieve the tag
 * @return         The tag value
 */
static const char *dav_dpm_getetag(const dav_resource *resource)
{
  return apr_psprintf(resource->pool, "%lx-%lx",
                      (unsigned long)resource->info->stat.fileid,
                      (unsigned long)resource->info->stat.mtime);
}

/* Hooks */
const dav_hooks_repository dav_hooks_repository_dpm =
{
    FLAG_HANDLE_GET,
    dav_dpm_get_resource,
    dav_dpm_get_parent,
    dav_dpm_is_same,
    dav_dpm_is_parent,
    dav_dpm_open_stream,
    dav_dpm_close_stream,
    dav_dpm_write_stream,
    dav_dpm_seek_stream,
    dav_dpm_set_headers,
    dav_dpm_deliver,
    dav_dpm_create_collection,
    dav_dpm_copy,
    dav_dpm_move,
    dav_dpm_remove,
    dav_dpm_walk,
    dav_dpm_getetag
};
