#include "mod_dav_dpm.h"
#include "mod_ssl-private.h"
#include <apr_strings.h>
#include <apr_tables.h>
#include <Cns_api.h>
#include <errno.h>
#include <http_log.h>
#include <serrno.h>
#include <stdarg.h>
#include <sys/stat.h>

/**
 * Replace all ocurrences of o with r
 * @param str The string where to perform the replacement
 * @param o   The character to be replaced
 * @param r   The character to put instead
 */
static char *replace(char *str, char o, char r)
{
  char *p;
  for(p = str; *p != '\0'; ++p) {
    if (*p == o)
      *p = r;
  }
  return str;
}

/**
 * Create a dav error with proper error message and HTTP status
 * @param p         The pool used to allocate memory
 * @param err_no    The error code (usually, the value of errno or serrno)
 * @param http_code If specified, this will be used instead of mapping err_no
 * @param format    Additional string to print (format)
 * @return          A initialized dav_error instance. Must not be freed.
 */
DAV_DECLARE(dav_error*) dav_dpm_new_error(apr_pool_t *p, int err_no, int http_code, const char *format, ...)
{
  char  buffer[DPM_MAX_ERROR_LEN + 1];
  char *err_msg;

  /* Map err_no to HTTP status code, if needed */
  if (!http_code) {
    switch(err_no)
    {
      case ENOENT:       http_code = HTTP_NOT_FOUND          ; break;

      case EPERM:
      case SENOMAPFND:
      case EACCES:       http_code = HTTP_FORBIDDEN          ; break;

      case ENOTDIR:      http_code = HTTP_FORBIDDEN          ; break;
      case EEXIST:       http_code = HTTP_METHOD_NOT_ALLOWED ; break;

      case ENOSYS:       http_code = HTTP_NOT_IMPLEMENTED    ; break;
      case ENAMETOOLONG: http_code = HTTP_REQUEST_URI_TOO_LARGE ; break;

      case ENOSPC:       http_code = HTTP_INSUFFICIENT_STORAGE; break;

      case SECOMERR:
      case SETIMEDOUT:   http_code = HTTP_SERVICE_UNAVAILABLE; break;

      default:           http_code = HTTP_INTERNAL_SERVER_ERROR;
    }
  }

  /* Get description of err_no */
  sstrerror_r(err_no, buffer, DPM_MAX_ERROR_LEN);

  /* If extra information has been specified, append it */
  if (format) {
    va_list extra;
    char    *add;

    va_start(extra, format);
    add = apr_pvsprintf(p, format, extra);
    va_end(extra);

    err_msg = apr_pstrcat(p, buffer, " - ", add, NULL);
  } else {
    err_msg = buffer;
  }

  return  dav_new_error(p, http_code, 0, err_msg);
}

/**
 * Interacts with mod_gridsite through environment variables and initializes
 * an instance of dav_dpm_user_info
 * Everything is created using Apache pool, so don't bother freeing
 * Returns NULL on error, and sets errno
 * @param r      The request
 * @param s_conf The server configuration
 * @param d_conf The directory configuration
 * @return  A pointer to an initialized dav_dpm_user_info point
 */
dav_dpm_user_info *dav_dpm_get_user_info(request_rec *r, dav_dpm_server_conf *s_conf, dav_dpm_dir_conf *d_conf)
{
  dav_dpm_user_info   *info;
  int                  i;
  const char          *s       = NULL;
  char                *p       = NULL;
  SSLConnRec          *sslconn = NULL;
  SSL                 *ssl     = NULL;
  SSL_SESSION         *ssl_session  = NULL;
  dav_dpm_user_map_cache_entry *ucache_first, *ucache_entry;

  /* Recover if present in the notes */
  info = (dav_dpm_user_info*)apr_table_get(r->connection->notes, "dav_dpm_user_info");
  if (info) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "Reusing previous mapping for %s (%i)",
                  info->principal, info->uid);
    return info;
  } else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                  "First request of the connection. Need to do the mapping.");
  }

  /* Create user information
   * We use the connection pool here because it has to be persistent through
   * requests */
  info = apr_pcalloc(r->connection->pool, sizeof(dav_dpm_user_info));

  /* Get user information from the connection notes */
  for (i = 0; (s = apr_table_get(r->connection->notes, apr_psprintf(r->pool, "GRST_CRED_AURI_%d", i))); ++i) {
    if (strncmp(s, "dn:", 3) == 0 && !info->principal && (p = index(s, '/'))) {
      info->principal = replace(apr_pstrdup(r->connection->pool, p), '+', ' ');

    }
    if (strncmp(s, "fqan:", 5) == 0 && info->nfqans < MAX_FQANS && (p = index(s, '/'))) {
      info->fqans[info->nfqans++] = replace(apr_pstrdup(r->connection->pool, p), '+', ' ');
    }
  }

  /* If not found, try with request notes */
  if (!i) {
    for (i = 0; (s = apr_table_get(r->notes, apr_psprintf(r->pool, "GRST_CRED_AURI_%d", i))); ++i) {
      if (strncmp(s, "dn:", 3) == 0 && !info->principal && (p = index(s, '/'))) {
        info->principal = replace(apr_pstrdup(r->connection->pool, p), '+', ' ');

      }
      if (strncmp(s, "fqan:", 5) == 0 && info->nfqans < MAX_FQANS && (p = index(s, '/'))) {
        info->fqans[info->nfqans++] = replace(apr_pstrdup(r->connection->pool, p), '+', ' ');
      }
    }
  }

  /* i = 0, that means there is no user certificate */
  if (!i) {
    /* Try to fallback to the anon */
    if (d_conf->anon_user) {
      info->principal = apr_pstrdup(r->connection->pool, d_conf->anon_user);
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "No certificate presented. Falling back to %s",
                    d_conf->anon_user);
    }
    else {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "No certificate presented.");
      serrno = EACCES;
      return NULL;
    }
  }

  /* VOMS from first fqans if any */
  if(info->nfqans) {
    s = info->fqans[0] + 1;
    p = index(s, '/');
    info->voname = apr_pmemdup(r->connection->pool, s, p - s);
    info->voname[p - s] = '\0';
  }

  /* Recover map if in cache */
  sslconn = (SSLConnRec*)ap_get_module_config(r->connection->conn_config, &ssl_module);
  if (sslconn && (ssl = sslconn->ssl) && s_conf->session_cache_shm) {
    /* Get cache */
    ucache_first = (dav_dpm_user_map_cache_entry*)apr_shm_baseaddr_get(s_conf->session_cache_shm);

    /* Get session */
    ssl_session  = SSL_get_session(ssl);

    /* Search */
    for (i = 0, ucache_entry = ucache_first; i < s_conf->session_cache_size; ++ucache_entry, ++i) {
      if (memcmp(ucache_entry->session_id, ssl_session->session_id, ssl_session->session_id_length) == 0) {
        /* Hit! */
        info->uid  = ucache_entry->uid;
        memcpy(info->gids, ucache_entry->gids, MAX_FQANS * sizeof(gid_t));

        ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                      "Recovered user mapping from cache! %s (%i)",
                      info->principal, info->uid);
        /* Save for later */
        apr_table_setn(r->connection->notes, "dav_dpm_user_info", (const char*)info);
        return info;
      }
    }
  }

  /* Map ID */
  if(Cns_getidmap(info->principal,
                   info->nfqans, (const char**)info->fqans,
                   &(info->uid), info->gids))
  {
    /* Try again with anon */
    if (d_conf->anon_user) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG, 0, r,
                    "Could not authenticate %s. Falling back to %s",
                    info->principal, d_conf->anon_user);
      info->principal = apr_pstrdup(r->connection->pool, d_conf->anon_user);
      if(Cns_getidmap(info->principal,
                      info->nfqans, (const char**)info->fqans,
                      &(info->uid), info->gids)) {
        /* No way */
        return NULL;
      }
    }
    else {
      return NULL;
    }
  }
  
  /* Save for later (keep-alive connections) */
  apr_table_setn(r->connection->notes, "dav_dpm_user_info", (const char*)info); /* I am cheating here a bit :) */

  /* Save in the cache */
  if (ucache_first && ssl_session) {
    long cur_time = time(NULL);
    for (i = 0, ucache_entry = ucache_first; i < s_conf->session_cache_size; ++ucache_entry, ++i) {
      if (ucache_entry->expiration < cur_time) {
        memcpy(ucache_entry->gids, info->gids, MAX_FQANS);
        ucache_entry->uid          = info->uid;
        ucache_entry->expiration   = SSL_SESSION_get_time(ssl_session) + SSL_SESSION_get_timeout(ssl_session);
        memcpy(ucache_entry->session_id, ssl_session->session_id, SSL_MAX_SSL_SESSION_ID_LENGTH);
        break;
      }
    }
  }

  /* Done here */
  return info;
}

/**
 * Return a string with the date-time formated
 * @param buffer Where to put the formated date
 * @param maxlen The size of the buffer
 * @param tstamp The timestamp
 * @param format Format flag
 * @return       The length of the printed string
 */
int dav_dpm_format_datetime(char *buffer, size_t maxlen, time_t tstamp, unsigned format)
{
  struct tm tms;
  int       l;

  gmtime_r(&tstamp, &tms);

  switch (format) {
    case DAV_DPM_RFC2068:
      l = snprintf(buffer, maxlen,
                   "%s, %.2d %s %d %.2d:%.2d:%.2d GMT",
                   apr_day_snames[tms.tm_wday],
                   tms.tm_mday, apr_month_snames[tms.tm_mon ],
                   tms.tm_year + 1900,
                   tms.tm_hour, tms.tm_min, tms.tm_sec);
      break;
    case DAV_DPM_RFC3339:
    default:
      l = snprintf(buffer, maxlen,
                   "%d-%.2d-%.2dT%.2d:%.2d:%.2dZ",
                   tms.tm_year + 1900, tms.tm_mon + 1,  tms.tm_mday,
                   tms.tm_hour, tms.tm_min, tms.tm_sec);
  }

  return l;
}

/**
 * Generates a printable version of the filemode
 * @param str  Where to put the string
 * @param mode The mode
 * @return     str
 */
char *dav_dpm_mode_str(char *str, mode_t mode)
{
  str[10] = '\0';

  if (S_ISDIR(mode))
    str[0] = 'd';
  else if (S_ISLNK(mode))
    str[0] = 'l';
  else
    str[0] = '-';
  
  str[1] = (mode & S_IRUSR) ? 'r':'-';
  str[2] = (mode & S_IWUSR) ? 'w':'-';
  if (mode & S_ISUID)
    str[3] = (mode & S_IXUSR) ? 's':'S';
  else
    str[3] = (mode & S_IXUSR) ? 'x':'-';

  str[4] = (mode & S_IRGRP) ? 'r':'-';
  str[5] = (mode & S_IWGRP) ? 'w':'-';
  if (mode & S_ISGID)
    str[6] = (mode & S_IXGRP) ? 's':'S';
  else
    str[6] = (mode & S_IXGRP) ? 'x':'-';

  str[7] = (mode & S_IROTH) ? 'r':'-';
  str[8] = (mode & S_IWOTH) ? 'w':'-';
  if (mode & S_ISVTX)
    str[9] = (mode & S_IXOTH) ? 't':'T';
  else
    str[9] = (mode & S_IXOTH) ? 'x':'-';

  return str;
}

/**
 * Generates a printable version of the size
 * @param str  Where to put the string
 * @param max  Max length
 * @param size The size in bytes
 * @return     str
 */
char *dav_dpm_size_str(char *str, size_t max, size_t size)
{
  if (size > 1073741823)
    snprintf(str, max, "%.1fG", (float)size / 1073741824.0);
  else if (size > 1048575)
    snprintf(str, max, "%.1fM", (float)size / 1048576.0);
  else if (size > 1023)
    snprintf(str, max, "%.1fK", (float)size / 1024.0);
  else
    snprintf(str, max, "%zu", size);
  return str;
}

/**
 * Add a response to the walk resource
 * @param wres  The walk resource
 * @param error The error to add
 */
void dav_dpm_add_response(dav_walk_resource *wres, dav_error *error)
{
  dav_response *resp;

  resp = apr_pcalloc(wres->pool, sizeof(dav_response));
  resp->href = apr_pstrdup(wres->pool, wres->resource->uri);
  resp->status = error->status;

  resp->next = wres->response;
  wres->response = resp;
}

/**
 * Parses a query (everything after a ?) and initializes a table with it
 * @param pool  The pool where memory is allocated
 * @param query The query to parse
 * @return      A apr_table_t with key=>value mapping
 */
apr_table_t *dav_dpm_parse_query(apr_pool_t *pool, const char *query)
{
  apr_table_t *table;
  char        *copy, *last, *t, *p;

  table = apr_table_make(pool, 0);

  if (!query)
    return table;

  copy  = apr_pstrdup(pool, query);

  /* Split by & (fields) */
  t = apr_strtok(copy, "&", &last);
  while (t) {
    /* Split by = (values) */
    p = strchr(t, '=');
    /* Careful with malformed URLs */
    if (p == NULL)
      break;
    *p = '\0';
    ++p;
    /* Now, in t we have the field, and in p the value*/
    apr_table_set(table, t, p);
    /* Next */
    t = apr_strtok(NULL, "&", &last);
  }

  return table;
}

/**
 * Gets the location from a TURL. HTTP/S aware. It will asume HTTPS for
 * unrecognised protocols.
 * @param pool     The pool
 * @param turl     The TURL to split
 * @param location Where to put the parsed URL
 */
void dav_dpm_split_turl(apr_pool_t *pool, const char *turl, struct physical_loc *location)
{
  apr_uri_t uri;

  apr_uri_parse(pool, turl, &uri);

  strncpy(location->host, uri.hostname, CA_MAXHOSTNAMELEN);
  strncpy(location->path, uri.path,     CA_MAXPATHLEN);


  if (strcmp(uri.scheme, "http") == 0) {
    strncpy(location->protocol, "http", CA_MAXPROTOLEN);
    location->port = uri.port_str ? uri.port : 80;
  }
  else if (strcmp(uri.scheme, "https") == 0) {
    strncpy(location->protocol, "https", CA_MAXPROTOLEN);
    location->port = uri.port_str ? uri.port : 443;
  }
  else {
    strncpy(location->protocol, "https", CA_MAXPROTOLEN);
    location->port = 443;
  }
}



char ssl_used(conn_rec *connection)
{
  return ap_get_module_config(connection->conn_config, &ssl_module) != NULL;
}
